-- Fenestra; version 0.80(beta); 1998-09-27

-- 
-- Copyright (c) 1998 Franck Arnaud, http://www.altsoft.demon.co.uk/
-- 
-- This package is distributed under the Eiffel Forum Freeware License, 
-- version 1 (see file forum.txt for full text). The main points are:
-- 
-- "Permission is hereby granted, without written agreement and without 
--  license or royalty fees, to use, copy, modify and/or distribute this 
--  package, provided that: copyright notices are retained unchanged [...]
-- 
--  Permission is hereby also granted [...] to distribute binary programs 
--  which depend on this package, provided that: if the binary program 
--  depends on a modified version of this package, you must publicly release 
--  the modified version of this package [...].
-- 
--  THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT WARRANTY [...]"
-- 

indexing
	cluster: Fenestra, Graphics
	description: "Abstract graphic device"
	interface: abstract

deferred class GRAPHIC_DEVICE

inherit
	WIN_OBJECT

	WIN_API

	WIN_RUNTIME_FINAL

feature -- Access

	is_ready: BOOLEAN is
			-- Is the device ready to accept graphic command?
		deferred
		end

	prepare is
			-- Get the device ready for graphic commands.
		require
			invalid: not is_ready
		deferred
		ensure
			ready: is_ready
		end

	release is
			-- Invalidate graphic device.
		require
			ready: is_ready
		deferred
		ensure
			invalid: not is_ready
		end

feature {NONE}

	font: FONT
	win_font: WIN_FONT
			-- Current font.
		
	brush: WIN_BRUSH
			-- Current brush.
		
	pen: WIN_PEN
			-- Current pen.

	palette: WIN_PALETTE
			-- Current palette.

	mapping: GRAPHIC_MAPPING
			-- Current mapping mode.

	rop2: INTEGER
			-- ROP2 mode.

	back_color: INTEGER
			-- Current background colour.

feature -- Capabilities

	capabilities: DEVICE_CAPABILITIES
			-- Device capabilities.

	text_capabilities: TEXT_METRICS
			-- Text capabilities.

feature -- Coordinates conversions

	logical_to_device (pt: POINT): POINT is
			-- Convert logical to device coordinates.
		require
			valid_point: pt /= Void
			ready: is_ready
		do
			wrt.win_gdi_lp2dp (handle, pt.x, pt.y)
			!!Result.make_coord (wrt.win_lastint1, wrt.win_lastint2)
		ensure
			is_copy: Result /= Void
		end

	device_to_logical (pt: POINT): POINT is
			-- Convert device to logical coordinates.
		require
			valid_point: pt /= Void
			ready: is_ready
		do
			wrt.win_gdi_dp2lp (handle, pt.x, pt.y)
			!!Result.make_coord (wrt.win_lastint1, wrt.win_lastint2)
		ensure
			is_copy: Result /= Void
		end
	
feature -- Set drawing attributes

	reset is
			-- Reset attributes.
		local
			default_brush: BRUSH
			default_pen: PEN
			def_color: RGB_COLOR
		do
			!SYSTEM_FONT!font.make
			
			!COLOR_BRUSH!default_brush.make
			brush := default_brush.to_windows
			
			!SIMPLE_PEN!default_pen.make
			pen := default_pen.to_windows

			-- Don't change system palette.

			-- bitmap?
			-- region?

			!PAGE_MAPPING!mapping.make
			
			rop2 := wrt.R2_copypen

			!!def_color.white
			back_color := def_color.to_windows

			-- if ready
			if is_ready then
				initialise_attributes
			end
		end

	set_mapping (mode: GRAPHIC_MAPPING) is
			-- Set mapping mode.
		require
			valid_mode: mode /= Void
		do
			mapping := mode.duplicate
			
			if is_ready then
				mapping.set_mode (Current)
			end
		end

	set_mode (mode: GRAPHIC_MODE) is
			-- Set the current graphic mode, used to determine how the 
			-- current pen and brush are applied. Note that it does not 
			-- apply to text.
		require
			valid_mode: mode /= Void
			raster: capabilities.is_raster_display --?
		do
			rop2 := mode.to_windows
			
			if is_ready then
				wrt.win_gdi_rop2 (handle, rop2)
			end
		end

	set_background_color (clr: COLOR) is
			-- Set the current background color, used by the device 
			-- for text drawn with a background area, and to fill 
			-- gaps between some styled lines (using SIMPLE_PEN) 
			-- for instance.
		require
			valid: clr /= Void
		do
			back_color := clr.to_windows

			if is_ready then
				wrt.win_gdi_bkcolor (handle, back_color)
			end
		end

	set_font (ft: FONT) is
			-- Set current font.
		require
			valid_font: ft /= Void
		do
			font := ft.duplicate
						
			if is_ready then
				select_font
			end
		end

	set_brush (br: BRUSH) is
			-- Set current brush
		require
			valid_brush: br /= Void
		do
			if brush /= Void then -- free old one.
				brush.force_dispose
			end
			brush := br.to_windows

			if is_ready then
				select_object (brush)
			end
		end

	set_pen (pn: PEN) is
			-- Set current pen.
		require
			valid_pen: pn /= Void
		do
			if pen /= Void then -- free old one.
				pen.force_dispose
			end
			pen := pn.to_windows

			if is_ready then
				select_object (pen)
			end
		end

	set_palette (pal: PALETTE) is
			-- Set current palette.
		require
			valid_pal: pal /= Void
			possible: is_ready implies capabilities.has_palette
			good_size: --? is_ready implies pal.size <= capabilities.palette_size
		do
			if palette /= Void then
				--? unrealize
				palette.force_dispose
			end
			palette := pal.to_windows
			if is_ready then
				realize_palette
			end
		end

feature -- Clipping

	is_point_visible (pt: POINT): BOOLEAN is
			-- Is point in the current clipping area?
		require
			valid_point: pt /= Void
			ready: is_ready
		do
			Result := wrt.win_gdi_ptvisu (handle, pt.x, pt.y)
		end

	is_rectangle_visible (rect: RECTANGLE): BOOLEAN is
			-- Does this rectangle has an intersection with the 
			-- current clipping area?
		require
			valid_rectangle: rect /= Void
			ready: is_ready
		do
			Result := wrt.win_gdi_rectvisu (handle, rect.lower.x, rect.lower.y,
						rect.upper.x, rect.upper.y)
		end

feature -- Actions

	draw (obj: DRAWABLE) is
			-- Draw the drawable object.
		require
			valid_object: obj /= Void 
			ready: is_ready
		do
			obj.draw (Current)
		end

feature -- Text metrics

	text_extent (obj: DRAWABLE_TEXT): RECTANGLE is
			-- Return the size of the bounding rectangle of the 
			-- text, as if it was drawn now.
		require
			valid_object: obj /= Void 
			ready: is_ready
		do
			Result := obj.text_extent (Current)
		ensure
			is_copy: Result /= Void
			origin: Result.lower.is_origin
		end
	
	character_width (ch: CHARACTER): INTEGER is
			-- Get witdh of a single character, as if it was drawn now.
		require
			ready: is_ready
		do
			Result := wrt.win_gdi_charsize (handle, ch.code)
		end

feature -- Colour 

	get_point_color (pt: POINT): COLOR is
			-- Get pixel colour.
		require
			ready: is_ready
			valid: pt /= Void
			raster: capabilities.is_raster_display
		local
			wincol: INTEGER
		do
			wincol := wrt.win_gdi_getpixel (handle, pt.x, pt.y)
			if wincol /= wrt.Fcd_bad_color then
				!STANDARD_COLOR!Result.make (wincol)
			end
		ensure
			void_if_outside: -- color = Void if outside clipping rectangle
		end

feature {WINDOW} -- Palette

	has_palette: BOOLEAN is
			-- Does this device currently has a palette?
		do
			Result := palette /= Void
		end

	realize_palette is
			-- `Realize' current palette: the hardware palette is changed 
			-- to match the logical palette. The palette is automatically 
			-- realized when it is `set' on a ready device or on a call to 
			-- `prepare' (including prepartion before `when_drawn' is called).
		require
			has_palette: has_palette
		do
			-- Select & Realize current palette
			wrt.win_gdi_selectpal (handle, palette.handle)
		end

feature {NONE} -- Creation init

	initialise is
			-- Creation-time initialisation
		require
			notready: not is_ready
		do
			finalisation_register
			!!capabilities.make (Current)
			!!text_capabilities.make (Current)
			reset
		ensure
			notready: not is_ready
		end

feature {NONE} -- Graphic attributes management

	initialise_attributes is
		require
			ready: is_ready
		do
			select_font
			select_object (brush)
			select_object (pen)
			if palette /= Void then
				realize_palette
			end
			-- bitmap?
			-- region?
			mapping.set_mode (Current)

			wrt.win_gdi_rop2 (handle, rop2)
			wrt.win_gdi_bkcolor (handle, back_color)
			wrt.win_gdi_bkmode (handle, wrt.Transparent)
		end

	select_font is
		require
			ready: is_ready
		do
			if win_font /= Void then
				win_font.force_dispose
			end
			win_font := font.to_windows (Current)
			select_object (win_font)
		end

	select_object (obj: WIN_GDI_OBJECT) is
			-- Select the object on the device
		require
			valid_object: obj /= Void
		do
			wrt.win_gdi_select (handle, obj.handle)
		end

invariant
	valid_attrib: pen /= Void and brush /= Void and mapping /= Void
	valid_caps: capabilities /= Void and then text_capabilities /= Void
	
end -- class GRAPHIC_DEVICE
