-- Pylon: a foundation library; version 0.82(beta); 1998-01-04

-- 
-- Copyright (c) 1997 Franck Arnaud, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY FRANCK ARNAUD AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE TO ANY PARTY FOR 
-- ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
-- DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE CODE.
-- 

indexing
	cluster: Pylon, Time
	description: "Gregorian date"
	interface: client

class P_DATE

inherit
	HASHABLE
		undefine
			out, is_equal
		end
	P_TEXT_OBJECT
		undefine
			out, is_equal
		end
	P_DATE_CONST
		undefine
			out, is_equal
		end
	COMPARABLE
		redefine
			out, is_equal
		end

feature -- Comparison

	infix "<" (other : like Current) : BOOLEAN is
		do
			Result := year < other.year 
				or else (year = other.year and (month < other.month
					or else (month = other.month and day < other.day)))
		end

	is_equal(other : like Current) : BOOLEAN is
		do
			Result := year = other.year and month = other.month and day = other.day
		end

feature -- Hashable
	
	hash_code : INTEGER is
		do
			Result := year * 400 + month * 31 + day
		end


feature {PUBLIC_NONE} -- Text object

	object_to_string(tm : P_TEXT_OBJECT_MANAGER) is
		do
			tm.put_integer(year)
			tm.put_integer(month)
			tm.put_integer(day)
		end

	string_to_object(tm : P_TEXT_OBJECT_MANAGER) is
		do
			tm.get_integer
			year := tm.last_integer
			tm.get_integer
			month := tm.last_integer
			tm.get_integer
			day := tm.last_integer

			if not tm.in_success then
				year := 0 day := 0 month := 0
			end
		end

feature

	year : INTEGER
			-- Full gregorian year.

	month : INTEGER
			-- Month (1-12)
		
	day : INTEGER
			-- Day (1-31)

	day_of_year : INTEGER is
			-- Day of year.
		require
			valid: is_valid
		local
			i : INTEGER
		do
			from
				i := 1
			variant
				month - i
			until
				i >= month
			loop
				Result := Result + max_days(i)
				i := i + 1
			end
			Result := Result + day
		ensure
			valid: Result > 0 and Result <= 366
		end

	day_of_week : INTEGER is
			-- Day of week (0 = Sunday).
		require
			valid: is_valid
			modern: year > 1752 -- and not russia < 19xx ...
		local
			m, century_year, century : INTEGER
		do
			-- Calculate day of week using Zeller's algorithm
			century := year // 100
			century_year := year \\ 100
			m := month

			-- march = 3; jan = 13, feb = 14 of previous year.
			if m < 3 then
				m := m + 12

				if century_year > 0 then
					century_year := century_year - 1
				else
					century_year := 99
					century := century - 1
				end
			end
			
			-- Zeller formula.
			Result := (day + ((26 * (m + 1)) // 10)
						+ century_year + (century_year // 4) + (century // 4)
						- (2 * century)) \\ 7 - 1

			-- remainder to modulo
			if Result < 0 then
				Result := Result + 7
			end
		ensure
			valid: Result >= 0 and Result < 7
		end

feature
	
	set_year(yyyy : INTEGER) is
			-- Set year.
		require
			positive: yyyy > 0
		do
			year := yyyy
		ensure
			done: yyyy = year
		end

	set_month(mm : INTEGER) is
			-- Set month.
		require
			valid: mm >= 1 and mm <= 12
		do
			month := mm
		ensure
			done: mm = month
		end

	set_day(dd : INTEGER) is
			-- Set day.
		require
			valid: dd >= 1 and dd <= 31
		do
			day := dd
		ensure
			done: dd = day
		end

	set(yyyy, mm, dd : INTEGER) is
			-- Set date.
		require
			valid_date: is_valid_date(yyyy, mm, dd)
		do
			set_year(yyyy)
			set_month(mm)
			set_day(dd)
		ensure
			done: dd = day and mm = month and yyyy = year
			valid: is_valid
		end
			
feature -- Status
	
	is_valid_date(yyyy, mm, dd : INTEGER) : BOOLEAN is
			-- Is the given date valid.
		do
			Result := yyyy > 0 and mm > 0 and mm <= 12 and dd > 0 and dd <= max_days(mm)
		end

	is_valid : BOOLEAN is
			-- Is the current date valid?
		do 
			Result := is_valid_date(year, month, day)
		end

	is_leap_year : BOOLEAN is
			-- Is the current year a leap year?
		do
			Result := (year \\ 4 = 0) and (not (year \\ 100 = 0) or (year \\ 400 = 0)) 
		end

feature {NONE}
	
	max_days(mm : INTEGER) : INTEGER is
			-- Max days in a month.
		do
			inspect
				mm
			when 4,6,9,11 then
				Result := 30
			when 2 then
				if is_leap_year then
					Result := 29
				else
					Result := 28
				end
			else
				Result := 31
			end
		end

feature -- Output

	out : STRING is
		do
			if is_valid then
				Result := to_iso
			else
				!!Result.make(10)
				Result.copy("00000000")
			end
		end

	to_iso_long : STRING is
			-- Output date in verbose ISO format.
			-- eg: "1887-07-26"
		require
			ok: is_valid
		local
			fmt : P_FORMAT
		do
			!!fmt
			!!Result.make(0)
			Result.copy(fmt.int_to_str_min(year,4))
			Result.append_string(Iso_separator)
			Result.append_string(fmt.int_to_str_min(month,2))
			Result.append_string(Iso_separator)
			Result.append_string(fmt.int_to_str_min(day,2))
		ensure
			is_copy: Result /= Void
		end

	to_iso : STRING is
			-- Output date in compact ISO format.
			-- eg: "18870726"
		require
			ok: is_valid
		local
			fmt : P_FORMAT
		do
			!!fmt
			!!Result.make(0)
			Result.copy(fmt.int_to_str_min(year,4))
			Result.append_string(fmt.int_to_str_min(month,2))
			Result.append_string(fmt.int_to_str_min(day,2))
		ensure
			is_copy: Result /= Void
		end

	to_rfc : STRING is
			-- Output date in canonical RFC 822 format.
			-- eg: "26 Jul 1887", in English.
		require
			ok: is_valid
		local
			fmt : P_FORMAT
		do
			!!fmt
			!!Result.make(0)
			Result.copy(fmt.int_to_str_min(day,2)) --? "01", " 1", 1"?
			Result.append_string(" ")
			inspect month
			when 1 then Result.append_string(Rfc_january)
			when 2 then Result.append_string(Rfc_february)
			when 3 then Result.append_string(Rfc_march)
			when 4 then Result.append_string(Rfc_april)
			when 5 then Result.append_string(Rfc_may)
			when 6 then Result.append_string(Rfc_june)
			when 7 then Result.append_string(Rfc_july)
			when 8 then Result.append_string(Rfc_august)
			when 9 then Result.append_string(Rfc_september)
			when 10 then Result.append_string(Rfc_october)
			when 11 then Result.append_string(Rfc_november)
			when 12 then Result.append_string(Rfc_december)
			end
			Result.append_string(" ")
			Result.append_string(fmt.int_to_str_min(year,1))
		ensure
			is_copy: Result /= Void
		end
		
	to_european : STRING is
			-- Output date using usual numeric european format.
			-- eg: "26/07/1887"
		require
			ok: is_valid
		local
			fmt : P_FORMAT
		do
			!!fmt
			!!Result.make(0)
			Result.copy(fmt.int_to_str_min(day,2))
			Result.append_string(Euro_separator)
			Result.append_string(fmt.int_to_str_min(month,2))
			Result.append_string(Euro_separator)
			Result.append_string(fmt.int_to_str_min(year,4))
		ensure
			is_copy: Result /= Void
		end

	to_american : STRING is
			-- Output date using usual numeric american format.
			-- eg: "07/26/1887"
		require
			ok: is_valid
		local
			fmt : P_FORMAT
		do
			!!fmt
			!!Result.make(0)
			Result.copy(fmt.int_to_str_min(month,2))
			Result.append_string(Us_separator)
			Result.append_string(fmt.int_to_str_min(day,2))
			Result.append_string(Us_separator)
			Result.append_string(fmt.int_to_str_min(year,4))
		ensure
			is_copy: Result /= Void
		end

feature {NONE} -- String constants

	Iso_separator : STRING is "-"
	Euro_separator : STRING is "/"
	Us_separator : STRING is "/"

end
