/*
 * AlterSoft Fenestra Win32 runtime system
 * -----------------------------------------------------------------
 * Copyright (c) 1998 Franck Arnaud
 *
 * Win32 API wrappers for the windowing interface
 *
 * Eiffel compiler dependencies:
 *	UniversalWindowCallback
 *	ControlSubclass
 *	win_subclass_control
 *	win_init_notifier
 *	win_init_notifier_ex
 */

#include "fenestra.h"

/* NOTES
 *
 * Naming conventions:
 *
 * Functions exported to the Eiffel system are prefixed with the 
 * preprocessor symbol FRTAPI and usually start with `win_'.
 * Functions used internally by the library are normally are prefixed 
 * with `fenestra_'.
 *
 */
 
/* ======================================================================
 * PARAMETER PASSING
 * ======================================================================
 */

/* These utility functions are used when a runtime function needs to 
 * return more than one value to Eiffel.
 */

/* global */

static EINT int1,int2,int3,int4;
static EBOOL bool1;
static LPTSTR str1=NULL;

static LPTSTR voidstr=TEXT("");

/* string */

FRTAPI EPOINTER win_str_get(void) 
{
	NULL_EXCEPTION(str1,"win_str_get");

	return((EPOINTER) str1);
}

FRTAPI void win_str_free(void) 
{
	NULL_EXCEPTION(str1,"win_str_free");

	/* free if not void */
	if(str1 != voidstr)
		fenestra_free(str1);

	str1 = NULL;
}

/* reset string to empty string */

void fenestra_void_string(void)
{
	ASSERT_EXCEPTION(str1 == NULL,"fenestra_void_string");
	str1 = voidstr;
}

/* pass string pointer */

void fenestra_pass_string(LPTSTR istr) 
{
	ASSERT_EXCEPTION((str1 == NULL) || (str1 == voidstr), "fenestra_pass_string");
	str1 = istr;
}

/* copy string */

void fenestra_copy_string(LPTSTR istr) 
{
	int cnt;

	ASSERT_EXCEPTION((str1 == NULL) || (str1 == voidstr), "fenestra_set_string");

	cnt = (lstrlen(istr)+1)*sizeof(TCHAR);
	
	str1 = fenestra_malloc(cnt);
	fenestra_memcpy(str1,istr,cnt);
}

/* other parameters */

FRTAPI EBOOL win_lastbool1(void) {
	return(bool1);
}

FRTAPI EINT win_lastint1(void) {
	return(int1);
}

FRTAPI EINT win_lastint2(void) {
	return(int2);
}

FRTAPI EINT win_lastint3(void) {
	return(int3);
}

FRTAPI EINT win_lastint4(void) {
	return(int4);
}

void fenestra_set_int(EINT i1) {
	int1 = i1;
}

void fenestra_set_int2(EINT i1, EINT i2) {
	int1 = i1;
	int2 = i2;
}

void fenestra_set_int4(EINT i1, EINT i2, EINT i3, EINT i4) {
	int1 = i1;
	int2 = i2;
	int3 = i3;
	int4 = i4;
}

void fenestra_set_bool(EBOOL b) {
	bool1 = b;
}

/* ======================================================================
 * UTILITIES
 * ======================================================================
 */

/* global */

static int error_mutex=0;

/* Runtime exception 
 *  LPTSTR  orig : name.
 */

void fenestra_exception(LPTSTR orig)
{
	LPTSTR serror;

	if (error_mutex == 0) /* Avoid recursive error message */
	{
		error_mutex = 1;
		serror = fenestra_malloc((lstrlen(orig) + 40) * sizeof(TCHAR)); 

		wsprintf(serror,TEXT("Win32 API: %s:%ld"),orig,(long) GetLastError());
		MessageBox(NULL,serror,TEXT("Program Error"),MB_TASKMODAL | MB_ICONSTOP | MB_OK);

		fenestra_free(serror);
		error_mutex = 0;
	}
	else
	{
		MessageBeep(0xFFFFFFFF);
	}
}


/* Reset a memory block
 *  void *  ptr  : start of block
 *  int     len  : block lenght
 */

void fenestra_memset(void *start, int len)
{
	LPBYTE ptr;
	int i;

	for (ptr = (BYTE *) start, i=0; i<len; i++,ptr++ )
		*ptr=0;
}

/* Copy a memory block
 *  void *  dest : destination block
 *  void *  src  : source block
 *  int     len  : block lenght
 */

void fenestra_memcpy(void *dest, void *src, int len)
{
	LPBYTE d,s;
	int i;
	
	for (d=(LPBYTE) dest, s=(LPBYTE) src, i=0; i<len; i++,d++,s++)
		*d=*s;
}

/* Allocate a local memory block.
 *  DWORD   size   : Bytes to allocate
 *  LPVOID  Result : Pointer to allocated block
 */

LPVOID fenestra_malloc(DWORD a)
{
	LPVOID result;

	ZERO_EXCEPTION(a,"fenestra_malloc.1");

	result = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,a);
	NULL_EXCEPTION(result, "fenestra_malloc.2");

	return(result);
}

/* Free a local memory block.
 *  LPDVOID ptr    : Pointer to the memory block to free
 */

void fenestra_free(LPVOID a)
{
	BOOL fres;

	fres = HeapFree(GetProcessHeap(),0,a);
	FALSE_EXCEPTION(fres, "fenestra_free");
} 

FRTAPI ECHAR win_int2char(EINT i)
{
#ifdef UNICODE
#error "Implement"
#else
	return((ECHAR) (unsigned char) i);
#endif
}

/* Eiffel malloc interface */

FRTAPI EPOINTER win_mem_alloc(EINT size)
{
	return(W2E_POINTER(fenestra_malloc(E2W_INT(size))));
}

FRTAPI void win_mem_free(EPOINTER ptr) 
{
	fenestra_free(E2W_POINTER(ptr));
}

/* ========================================================================
 * NOTIFIER INTERFACE
 * ========================================================================
 */

/* This section contains all callbacks to the Eiffel system and is therefore 
 * heavily (Eiffel) compiler dependent.
 */

/* compiler dependencies */

#if defined(EIFFELS)
#define NOTIFIER_OBJ 			OBJREF							/* NOTIFIER object */
#define NOTIFIER_MESSAGE		notifier_message				/* callback functions */
#define NOTIFIER_MSG_KEY		notifier_msg_control_key
#define NOTIFIER_MSG_CTL		notifier_msg_control
#define PROTECT(a)				RGAC_protect(&a,1)				/* Keep object */
#define NOTIFIER_INSTANCE		ournot

#elif defined(TOWER) 
#define NOTIFIER_OBJ 			EIF_OBJ
#define NOTIFIER_MESSAGE		notifier__message
#define NOTIFIER_MSG_KEY		notifier__msg_control_key
#define NOTIFIER_MSG_CTL		notifier__msg_control
#define PROTECT(a) 				eif_adopt(a)
#define NOTIFIER_INSTANCE		ournot

#elif defined(CECIL)
#define NOTIFIER_OBJ 			void *
#define NOTIFIER_MESSAGE		(*notif_message)
#define NOTIFIER_MSG_KEY		(*notif_msg_control_key)
#define NOTIFIER_MSG_CTL		(*notif_msg_control)
#define PROTECT(a) 				eif_adopt(a)
#define NOTIFIER_INSTANCE		eif_access(ournot)

static EIF_PROC notif_message;
static EIF_FN_BOOL notif_msg_control_key;
static EIF_PROC notif_msg_control;

#elif defined(SMALLEIFFEL)
#define NOTIFIER_OBJ 			void *							/* NOTIFIER object */
#define NOTIFIER_MESSAGE		notifier_message				/* callback functions */
#define NOTIFIER_MSG_KEY		notifier_msg_control_key
#define NOTIFIER_MSG_CTL		notifier_msg_control
#define PROTECT(a)				UNREFERENCED_PARAMETER(a) /* Keep object anyway (boehm gc) */
#define NOTIFIER_INSTANCE		ournot

#elif defined(ANY_EIFFEL) /* generic support using callbacks */
#define NOTIFIER_OBJ			LPVOID
#define NOTIFIER_MESSAGE		(*notif_message)
#define NOTIFIER_MSG_KEY		(*notif_msg_control_key)
#define NOTIFIER_MSG_CTL		(*notif_msg_control)
#define PROTECT(a) 				UNREFERENCED_PARAMETER(a) /* notifier is once anyway */
#define NOTIFIER_INSTANCE		ournot

/* function pointers to callbacks */
static void (*notif_message)(LPVOID,E_HWND,EINT,ECHAR,EINT,EINT,EBOOL);
static EBOOL (*notif_msg_control_key)(LPVOID,E_HWND,ECHAR);
static void (*notif_msg_control)(LPVOID,E_HWND,EINT,ECHAR);

/* DLL main */
BOOL WINAPI FenestraDllMain(HINSTANCE inst, DWORD reason, LPVOID reserved)
{
	/* nothing special to do */
	return(TRUE);
}

#endif

/* ------------------------------------------------------------------
 * NOTIFIER (callbacks) interface
 */

/* local global */
static NOTIFIER_OBJ ournot; /* the notifier */

/* notifier init */

FRTAPI void win_init_notifier(NOTIFIER_OBJ current)
{
	/* register ourselves for future call(backs) */
#if defined(CECIL) /* initialise cecil function pointer */
	EIF_TYPE_ID notid;
	
#ifdef ISSCECIL
	ournot = eif_adopt (current);
#else
	ournot = current;
	PROTECT(ournot);
#endif
	
	notid = eif_type_id("NOTIFIER");
	ASSERT_EXCEPTION(notid != EIF_NO_TYPE,"win_init_notifier.1");

	notif_message = eif_proc("message",notid);
	/*ASSERT_EXCEPTION(notif_message != EIF_NO_ROUTINE,"win_init_notifier.2");*/
	notif_msg_control_key = eif_fn_bool("msg_control_key",notid);
	/*ASSERT_EXCEPTION(notif_msg_control_key != EIF_NO_ROUTINE,"win_init_notifier.3");*/
	notif_msg_control = eif_proc("msg_control",notid);
	/*ASSERT_EXCEPTION(notif_msg_control != EIF_NO_ROUTINE,"win_init_notifier.4");*/
#else
	ournot = current;
	PROTECT(current); /* useful? */
#endif
}

FRTAPI void win_init_notifier_ex(NOTIFIER_OBJ current, EPOINTER msg_func, EPOINTER msg_ctl, EPOINTER msg_key)
{
#ifdef ANY_EIFFEL
	ournot = current;
	notif_message = E2W_POINTER(msg_func);
	notif_msg_control = E2W_POINTER(msg_ctl);
	notif_msg_control_key = E2W_POINTER(msg_key);
#else
	UNREFERENCED_PARAMETER(current);
	UNREFERENCED_PARAMETER(msg_func);
	UNREFERENCED_PARAMETER(msg_ctl);
	UNREFERENCED_PARAMETER(msg_key);
#endif
}

/* ------------------------------------------------------------------
 * control subclassing 
 */

/* subclass builtin control */

FRTAPI void win_subclass_control(E_HWND win)
{
	LONG oldwp;
	BOOL fres;

	oldwp = SetWindowLong(E2W_HWND(win),GWL_WNDPROC,(LONG) ControlSubclass);
	ZERO_EXCEPTION(oldwp, "win_subclass_control.1"); 

	fres = SetProp(E2W_HWND(win),PROP_CALLBACK,(HANDLE)oldwp);
	FALSE_EXCEPTION(fres, "win_subclass_control.2");
}

/* Subclassed callback for builtin controls */

LRESULT CALLBACK ControlSubclass(HWND wid, UINT command, WPARAM wparam, LPARAM lparam)
{
#ifndef NO_NOTIFIER
	WNDPROC wp;
	ECHAR echar;
	EINT ecode;
	E_HWND ewin;
	LONG fres1;
	HANDLE fres2;
	
	wp = (WNDPROC) GetProp(wid,PROP_CALLBACK);
	NULL_EXCEPTION(wp,"ControlSubclass");

	ewin = W2E_HWND(wid);
	ecode = W2E_INT(command);
	echar = W2E_CHAR('\0');
	
	if(command==WM_KEYDOWN)
	{
		if (wparam == VK_UP)
		{
			echar = W2E_CHAR('\1');
			if(NOTIFIER_MSG_KEY(NOTIFIER_INSTANCE,ewin,echar) == ETRUE)
				return(0);
		}
		else if (wparam == VK_DOWN)
		{	
			echar = W2E_CHAR('\2');
			if(NOTIFIER_MSG_KEY(NOTIFIER_INSTANCE,ewin,echar) == ETRUE)
				return(0);
		}
	}
	else if(command==WM_CHAR)
	{
		echar = W2E_CHAR(wparam);
		if(NOTIFIER_MSG_KEY(NOTIFIER_INSTANCE,ewin,echar) == ETRUE)
			return(0);
	}
	else if(command==WM_SYSCHAR) /* Alt-<key>, used by dialog keyboard interface */
	{
		echar=W2E_CHAR(wparam);
		if((echar>=32) || (echar < 0))
		{
			NOTIFIER_MSG_CTL(NOTIFIER_INSTANCE,ewin,ecode,echar);
			return(0);
		}
	}
	else if((command==WM_SETFOCUS) || (command==WM_KILLFOCUS))
	{
		NOTIFIER_MSG_CTL(NOTIFIER_INSTANCE,ewin,ecode,echar);
	}
	else if(command==WM_DESTROY)
	{
		NOTIFIER_MSG_CTL(NOTIFIER_INSTANCE,ewin,ecode,echar); /* unregister from notifier */

		fres1=SetWindowLong(wid,GWL_WNDPROC,(LONG) wp);	/* cancel subclass */
		ZERO_EXCEPTION(fres1,"ControlSubclass.1");

		fres2=RemoveProp(wid,PROP_CALLBACK); /* remove property */
		NULL_EXCEPTION(fres2,"ControlSubclass.2");
	}

	return(CallWindowProc(wp,wid,command,wparam,lparam));
#endif /* !NO_NOTIFIER */
}


/* ------------------------------------------------------------------
 * Universal window callback
 */ 


LRESULT CALLBACK UniversalWindowCallback(HWND wid, UINT command, WPARAM wparam, LPARAM lparam)
{
#ifndef NO_NOTIFIER
	HWND event_win;
	E_HWND ewin;
	
	EINT event_code, event_int1, event_int2;
	ECHAR event_char;
	EBOOL queueable;

	/* reformat parameters */
	event_win=(HWND) NULL;

	event_code = W2E_INT(command);
	event_int1 = event_int2 = W2E_INT(0);
	event_char = W2E_CHAR((TCHAR) 0);
	queueable = TRUE;

	switch(command)
	{
		case WM_PAINT:
		case WM_QUERYNEWPALETTE: /* fixme: return TRUE if changed? */
		case WM_CLOSE:
		case WM_DESTROY:
		case WM_SETFOCUS:
		case WM_KILLFOCUS:
		case WM_LBUTTONUP:
		case WM_RBUTTONUP:
		event_win = wid;							/* window */
		break;

		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONDBLCLK:
		case WM_MOUSEMOVE:
		event_int1 = W2E_INT(((SHORT)LOWORD(lparam)));	/* x coord */
		event_int2 = W2E_INT(((SHORT)HIWORD(lparam)));	/* y coord */
		event_win = wid;								/* window */
		break;

		case WM_KEYDOWN:
		case WM_KEYUP:
		case WM_DROPFILES:
		case WM_TIMER:
		case WM_SIZE:
		case WM_INITMENU:
		case WM_INITMENUPOPUP:
		event_int1 = W2E_INT(wparam);	   			/* virtual key / drop handle / menu handle */
		event_win = wid;							/* window */
		break;

		case WM_CHAR:
		event_char = W2E_CHAR(((TCHAR) wparam));	/* character */
		event_win = wid;					   		/* window */
		break;	
		
		case WM_HSCROLL:
		case WM_VSCROLL:
		if(lparam) /* detect control (not clearly documented method) */
		{
			event_code = W2E_INT(FILO_SCROLL);	
			event_win = ((HWND) lparam);			/* window (control) */
		}
		else /* built into child window */		
		{
			event_win = wid;						/* window (child) */
		}											
		event_int1 = W2E_INT(LOWORD(wparam));		/**/
		event_int2 = W2E_INT(HIWORD(wparam));		/**/
		
		queueable = W2E_BOOL(FALSE); /* for live scrolling */
		break;	

		case WM_COMMAND:
		if(lparam)
		{ /* control */
			event_code = W2E_INT(FILO_COMMAND);
			event_int1 = W2E_INT(HIWORD(wparam));		/* notification code */
			event_win = ((HWND) lparam);				/* window */
		}
		else
		{ /* menu */
			event_int1 = W2E_INT(LOWORD(wparam));		/* menu ID */
			event_win = wid;							/* window */
		}
		break;
	
		case WM_MENUSELECT:
		if(HIWORD(wparam) == 0xFFFF)
		{
			event_int1 = 0;
			event_int2 = 1;
			event_win = wid;
		}
		else
		{
			if((HIWORD(wparam) & (MF_POPUP|MF_SYSMENU)) == 0) /* if not popup */
			{ 
				event_int1 = W2E_INT(LOWORD(wparam)); /* ID */
				event_int2 = 0;
				event_win = wid;
			}
		}
		break;
	}

	if(event_win != NULL)
	{
		/* callback eiffel */
		ewin = W2E_HWND(event_win);

		NOTIFIER_MESSAGE(NOTIFIER_INSTANCE, ewin, event_code, event_char, event_int1, event_int2, queueable);
		return(0);
	}

	return(DefWindowProc(wid,command,wparam,lparam));
#endif /* !NO_NOTIFIER */
}


/* -----------------------------------------------------------------------
 * Font enumeration
 */

struct frt_enum_font /* (internal) structure used for temporary data during the callback */
{
	int current_index; /* current index where next entry will be written (in LOGFONT) */
	int buffer_size; /* buffer size (in LOGFONT) */
	LOGFONT *buffer; /* buffer of LOGFONT structures. size >= sizeof(LOGFONT) * buffer_size */

	int fonts_count; /* total number of fonts available (callbacks count) */
};

int CALLBACK EnumFontFamCallback(ENUMLOGFONT *lpelf, NEWTEXTMETRIC FAR *lpntm, int FontType, LPARAM lParam)
{
	struct frt_enum_font *callback_struct;

	callback_struct = (struct frt_enum_font *) lParam; /* convert */

	/* if not full, copy the LOGFONT structure */
	if (callback_struct->current_index < callback_struct->buffer_size)
	{
		/* copy the LOGFONT that is at the beginning of ENUMLOGFONT to our LOGFONT buffer */
		fenestra_memcpy (callback_struct->buffer + callback_struct->current_index, lpelf, sizeof (LOGFONT));
		(callback_struct->current_index)++;
	}

	(callback_struct->fonts_count)++;
	return(1); /* continue enumeration */
}

FRTAPI EINT win_gdi_enumfonts(E_HGDI dcid, ESTRING typename, EINT logfont_buf_size, EPOINTER logfont_buf)
{
	LPTSTR tfname;
	struct frt_enum_font callback_struct;

	callback_struct.current_index = 0;
	callback_struct.buffer_size = E2W_INT (logfont_buf_size);
	callback_struct.buffer = (LOGFONT *) E2W_POINTER (logfont_buf);
	callback_struct.fonts_count = 0;

	tfname = E2W_STRING(typename);
	if (*tfname==0)
		tfname = NULL;

	EnumFontFamilies (E2W_HGDI(dcid), tfname, EnumFontFamCallback, (LPARAM) &callback_struct);
	/* result: last value returned by callback (?) */

	return(W2E_INT (callback_struct.fonts_count));
}


/* ================================================================
 * WINDOW MANAGEMENT
 * ================================================================
 */

/* global variables */

static HWND acc_window = NULL;
static HACCEL acc_table;

/* process one event */

FRTAPI EBOOL win_event(void )
{
	MSG msg;
	
	if( GetMessage( &msg, NULL, 0, 0 ) == TRUE)	
	{
		if((acc_window == NULL) || (!TranslateAccelerator(acc_window,acc_table,&msg)))
		{
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
		return(ETRUE);
	}
	return(EFALSE);
}

/* enable accelerators */

FRTAPI void win_accel_enable(E_HWND target_win, E_HWND accel_table)
{
	acc_window = E2W_HWND(target_win);
	acc_table = (HACCEL) E2W_HWND(accel_table);
}

/* register our class(es) */

FRTAPI void win_initwin(void )
{		
	WNDCLASS wc;
	ATOM fres;

	/* register our window class(es) */
	wc.style = CS_DBLCLKS; /* DC? CS_OWNDC Redraw? CS_HREDRAW | CS_VREDRAW */
	wc.lpfnWndProc = UniversalWindowCallback;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = PROGHINST;
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
	NULL_EXCEPTION(wc.hbrBackground,"win_initwin.1");
	
	wc.lpszMenuName = NULL;
	wc.lpszClassName = CLASS_OVERLAPPED;
	fres=RegisterClass(&wc);
	ASSERT_EXCEPTION(fres != 0,"win_initwin.2");

	wc.hbrBackground = NULL;
	wc.lpszClassName = CLASS_CHILDWINDOW;
	fres=RegisterClass(&wc);
	ASSERT_EXCEPTION(fres != 0,"win_initwin.3");

#if (WINVER >= 0x0400)
	wc.hbrBackground = (HBRUSH) (COLOR_3DFACE + 1);
#else
	wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
#endif
	wc.hIcon = NULL;
	wc.lpszClassName = CLASS_DIALOG;
	fres=RegisterClass(&wc);
	ASSERT_EXCEPTION(fres !=0,"win_initwin.3");

	wc.hCursor = NULL;
	wc.lpszClassName = CLASS_FREECURSOR;
	fres=RegisterClass(&wc);
	ASSERT_EXCEPTION(fres !=0,"win_initwin.4");
}

/* create overlapped window */

FRTAPI E_HWND win_create( void )
{
	HWND win;

	win = CreateWindowEx(0,CLASS_OVERLAPPED,TEXT("Fenestra Window"),WS_OVERLAPPEDWINDOW,
				CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,
				NULL, NULL, PROGHINST, NULL);
	NULL_EXCEPTION(win,"win_create");
	
	return(W2E_HWND(win));
}						 

/* create popup window */

FRTAPI E_HWND win_create_popup(E_HWND parent)
{
	HWND win;

	win = CreateWindowEx(0,TEXT(CLASS_OVERLAPPED),TEXT("Fenestra Popup"),
			WS_POPUPWINDOW | WS_CAPTION,
			0,0,0,0,
			/*CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,*/
			E2W_HWND(parent), NULL, PROGHINST, NULL);
	NULL_EXCEPTION(win,"win_create_popup");

	return(W2E_HWND(win));
}

/* create dialog window */

static void win_init_dialog(HWND win)
{
	HMENU sysm;
	int i,nitem;
	UINT mstate;
	
	/* update the system menu */
	sysm=GetSystemMenu(win,FALSE);
	NULL_EXCEPTION(sysm,"win_init_dialog");

	DeleteMenu(sysm, SC_RESTORE, MF_BYCOMMAND); /* fixme: cosmetic, so don't check success */
	DeleteMenu(sysm, SC_MINIMIZE, MF_BYCOMMAND);
	DeleteMenu(sysm, SC_MAXIMIZE, MF_BYCOMMAND);
	DeleteMenu(sysm, SC_SIZE, MF_BYCOMMAND);
	
	for(i=0,nitem=GetMenuItemCount(sysm);i<nitem;i++) /* nitem = -1 if error */
	{
		mstate=GetMenuState(sysm,i,MF_BYPOSITION);
		if((mstate != 0xFFFFFFFF) && (mstate & MF_SEPARATOR))
		{
			DeleteMenu(sysm,i,MF_BYPOSITION);
			break; /* for. only first one */
		} 
	}
}

FRTAPI E_HWND win_create_dialog(E_HWND parent, EBOOL thick)
{
	HWND win;

	win = CreateWindowEx((thick == ETRUE) ? WS_EX_DLGMODALFRAME : 0,
			TEXT(CLASS_DIALOG),TEXT("Fenestra Dialog"),
			WS_POPUPWINDOW | WS_DLGFRAME | WS_CAPTION,
			0,0,0,0,
			E2W_HWND(parent), NULL, PROGHINST, NULL);
	NULL_EXCEPTION(win,"win_create_dialog");

	win_init_dialog(win);
	return(W2E_HWND(win));
}

FRTAPI E_HWND win_create_appdialog(EBOOL thick)
{
	HWND win;

	win = CreateWindowEx((thick == ETRUE) ? WS_EX_DLGMODALFRAME : 0,
			TEXT(CLASS_DIALOG),TEXT("Fenestra Dialog"),
			WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION,
			CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,
			NULL, NULL, PROGHINST, NULL);
	NULL_EXCEPTION(win,"win_create_appdialog");

	win_init_dialog(win);
	return(W2E_HWND(win));
}


/* create child window */

FRTAPI E_HWND win_create_child(E_HWND parent, EBOOL cursor, EBOOL horscroll, EBOOL verscroll)
{
	HWND win;
	DWORD style;

	style = WS_CHILD | WS_VISIBLE | WS_BORDER;

	if(horscroll == ETRUE) 
		style |= WS_HSCROLL;
	if(verscroll == ETRUE)
		style |= WS_VSCROLL;
	
	win = CreateWindowEx(0,cursor == ETRUE ? TEXT(CLASS_FREECURSOR) : TEXT(CLASS_CHILDWINDOW),TEXT("Fenestra Child"),
			style,
			0,0,0,0,
			E2W_HWND(parent), NULL, PROGHINST, NULL);
	NULL_EXCEPTION(win,"win_create_child");
	
	return(W2E_HWND(win));
}

FRTAPI E_HWND win_create_clrchild(E_HWND parent, EINT color, EBOOL cursor, EBOOL horscroll, EBOOL verscroll)
{
	WNDCLASS wc;
	TCHAR classname[100];
	ATOM fres;
	HWND win;
	COLORREF wcolor;
	DWORD style;
	LOGBRUSH br;
	HBRUSH brush;

	/* color */
	wcolor = (COLORREF) E2W_INT(color);

	/* create class */
	wsprintf(classname,"%s%x",TEXT(CLASS_CHILDWINDOW),(DWORD) E2W_INT(color));
	if(GetClassInfo(PROGHINST,classname,&wc)!=TRUE)
	{
		/* prepare brush */
		br.lbStyle = BS_SOLID;
		br.lbColor = wcolor;
		br.lbHatch = 0;

		brush=CreateBrushIndirect(&br);
		NULL_EXCEPTION(brush,"win_create_clrchild.1");

		/* register our window class(es) */
		wc.style = CS_DBLCLKS; /* DC? CS_OWNDC Redraw? CS_HREDRAW | CS_VREDRAW */
		wc.lpfnWndProc = UniversalWindowCallback;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = PROGHINST;
		wc.hIcon = NULL;
		wc.hCursor = (cursor==ETRUE) ? NULL : LoadCursor(NULL,IDC_ARROW);
		wc.hbrBackground = brush;
		wc.lpszMenuName = NULL;
		wc.lpszClassName = classname;
	
		fres=RegisterClass(&wc);
		ASSERT_EXCEPTION(fres!=0,"win_create_clrchild.2")
	}

	/* style */
	style = WS_CHILD | WS_VISIBLE | WS_BORDER;
	if(horscroll == ETRUE) 
		style |= WS_HSCROLL;
	if(verscroll == ETRUE)
		style |= WS_VSCROLL;
	
	/* actual creation */
	win = CreateWindowEx(0,classname,TEXT("Fenestra Child"),
			style,
			0,0,0,0,
			E2W_HWND(parent), NULL, PROGHINST, NULL);
	NULL_EXCEPTION(win,"win_create_clrchild.3");

	return(W2E_HWND(win));
}

/* border */

FRTAPI void win_border(E_HWND ewin, EBOOL newborder)
{
	LONG style,fres;
	HWND win;

	win = E2W_HWND(ewin);
	style = GetWindowLong(win,GWL_STYLE);
	
	ASSERT_EXCEPTION(style!=0, "win_border.1"); /* WS_CHILD can't be 0 */

	style &= ~ WS_BORDER; /* remove old border bit */
	if(newborder == ETRUE)
		style |= WS_BORDER;

	fres=SetWindowLong(win,GWL_STYLE,style);
	ZERO_EXCEPTION(fres, "win_border.2");
}

/* show window */

FRTAPI void win_show(E_HWND win, EINT how)
{
	ShowWindow(E2W_HWND(win), E2W_INT(how));
	/* return value irrelevant */
}

/* enable window */

FRTAPI void win_enable(E_HWND win)
{
	EnableWindow(E2W_HWND(win), TRUE);
	/* return value irrelevant */ 
}

/* disable window */

FRTAPI void win_disable(E_HWND win)
{
	EnableWindow(E2W_HWND(win), FALSE);
	/* return value irrelevant */
}

/* focus */

FRTAPI void win_focus(E_HWND win)
{
	SetFocus(E2W_HWND(win)); 
	/* return value irrelevant */
}

FRTAPI EBOOL win_hasfocus(E_HWND win)
{
	if(GetFocus()==E2W_HWND(win))
		return(ETRUE);
	return(EFALSE);
}

/* activate */

FRTAPI void win_activate(E_HWND win)
{
	SetActiveWindow(E2W_HWND(win));
	/* return value irrelevant */
}

FRTAPI EBOOL win_isactive(E_HWND win)
{
	if(GetActiveWindow()==E2W_HWND(win))
		return(ETRUE);
	return(EFALSE);
}

/* update window */

FRTAPI void win_update(E_HWND win)
{
	BOOL fres;
	
	fres = InvalidateRect(E2W_HWND(win),NULL,TRUE);
	FALSE_EXCEPTION(fres,"win_update");
}

FRTAPI void win_update_rect(E_HWND win, EINT x, EINT y, EINT xp, EINT yp)
{
	RECT r;
	BOOL fres;

	SetRect(&r,E2W_INT(x),E2W_INT(y),E2W_INT(xp),E2W_INT(yp)); 
	fres=InvalidateRect(E2W_HWND(win),&r,TRUE);
	FALSE_EXCEPTION(fres,"win_update_rect");
}

FRTAPI void win_scroll(E_HWND win, EINT dx, EINT dy)
{
	int fres;

	fres = ScrollWindowEx(E2W_HWND(win),E2W_INT(dx),E2W_INT(dy),
			NULL,NULL,NULL,NULL,SW_INVALIDATE);
	ASSERT_EXCEPTION(fres != ERROR, "win_scroll");
}

/* get window size, returns 4 static ints */

FRTAPI void win_size(E_HWND win, E_HWND parent)
{
	BOOL fres;
	RECT r;
	POINT p1,p2;

	fres=GetWindowRect(E2W_HWND(win),&r);
	FALSE_EXCEPTION(fres,"win_size");
	
	p1.x = r.left;
	p1.y = r.top;
	if(parent) ScreenToClient(E2W_HWND(parent),&p1); /* fixme: should check? */
	p2.x = r.right;
	p2.y = r.bottom;
	if(parent) ScreenToClient(E2W_HWND(parent),&p2); /* fixme: should check? */

	fenestra_set_int4(W2E_INT(p1.x), W2E_INT(p1.y), W2E_INT(p2.x), W2E_INT(p2.y));
}

/* get window client size */

FRTAPI void win_clientsize(E_HWND win)
{
	BOOL fres;
	RECT r;

	fres=GetClientRect(E2W_HWND(win),&r);
	FALSE_EXCEPTION(fres,"win_clientsize");
	
	fenestra_set_int4(W2E_INT(r.left), W2E_INT(r.top), W2E_INT(r.right), W2E_INT(r.bottom));
}

/* resize */

FRTAPI void win_resize(E_HWND win,EINT x,EINT y,EINT xp,EINT yp)
{
	BOOL fres;

	fres = MoveWindow(E2W_HWND(win), E2W_INT(x), E2W_INT(y),
							  E2W_INT(xp - x), E2W_INT(yp - y), TRUE);
	FALSE_EXCEPTION(fres, "win_resize");
}

/* resize client */

FRTAPI void win_resizeclient(E_HWND win,EINT x,EINT y,EINT xp,EINT yp)
{
	RECT old,r;
	BOOL hasmenu,fres;

	/* get old window rectangle [for (x,y)] */
	fres=GetWindowRect(E2W_HWND(win),&old);
	FALSE_EXCEPTION(fres, "win_resizeclient.1");

	/* prepare rectangle of client area */
	SetRect(&r,E2W_INT(x),E2W_INT(y),E2W_INT(xp),E2W_INT(yp)); 
	/* menu flag */
	hasmenu = GetMenu(E2W_HWND(win))== NULL ? FALSE : TRUE;
	/* convert to size */
	fres=AdjustWindowRectEx(&r,(DWORD)GetWindowLong(E2W_HWND(win),GWL_STYLE),hasmenu,
							(DWORD)GetWindowLong(E2W_HWND(win),GWL_EXSTYLE));
	FALSE_EXCEPTION(fres,"win_resizeclient.2");
	
	fres=MoveWindow(E2W_HWND(win),old.left,old.top,r.right - r.left,r.bottom - r.top,TRUE);
	FALSE_EXCEPTION(fres,"win_resizeclient.3"); 
}

/* mouse capture on */

FRTAPI void win_capture(E_HWND win)
{
	SetCapture(E2W_HWND(win));
	/* return value irrelevant */
}

/* mouse capure off */

FRTAPI void win_endcapture(void )
{
	BOOL fres;

	fres=ReleaseCapture();
	FALSE_EXCEPTION(fres,"win_endcapture");
}

/* set text */

FRTAPI void win_title(E_HWND win, ESTRING string)
{
	BOOL fres;

	fres=SetWindowText(E2W_HWND(win), E2W_STRING(string));
	FALSE_EXCEPTION(fres,"win_title");
}

/* get text */

FRTAPI void win_gettitle(E_HWND win)
{
	int textlen;
	LPTSTR wintext;

	fenestra_void_string();

	textlen=GetWindowTextLength(E2W_HWND(win));
	if(textlen > 0)
	{
		wintext=fenestra_malloc((textlen + 2) * sizeof(TCHAR));
		
		textlen=GetWindowText(E2W_HWND(win),wintext,textlen+1);
		if(textlen<=0) {
			wintext[0]=((TCHAR) 0);
		}

		fenestra_pass_string(wintext); /* will free */
	}
}

/* close/minize window */

FRTAPI void win_close(E_HWND win)
{
	BOOL fres;

	fres=CloseWindow(E2W_HWND(win));
	FALSE_EXCEPTION(fres,"win_close");
}

/* destroy window */

FRTAPI void win_destroy(E_HWND win)
{
	BOOL fres;

	fres=DestroyWindow(E2W_HWND(win));
	FALSE_EXCEPTION(fres,"win_destroy");
}

/* QUIT */

FRTAPI void win_quit(void )
{
	PostQuitMessage(0);
}

/* queries */

FRTAPI EBOOL win_isicon(E_HWND win)
{
	if(IsIconic(E2W_HWND(win))==TRUE)
		return(ETRUE);

	return(EFALSE);
}

FRTAPI EBOOL win_isenable(E_HWND win)
{
	if(IsWindowEnabled(E2W_HWND(win))==TRUE)
		return(ETRUE);

	return(EFALSE);
}

FRTAPI EBOOL win_isvisible(E_HWND win)
{
	if(IsWindowVisible(E2W_HWND(win))==TRUE)
		return(ETRUE);

	return(EFALSE);
}

/* key status */

FRTAPI EBOOL win_keystate(EINT vkey)
{
	SHORT ret;

	ret=GetAsyncKeyState(E2W_INT(vkey));
	if(ret & 0x8000)
		return(ETRUE);
	return(EFALSE);
}

/* window font */

FRTAPI void win_setfont(E_HWND win, E_HGDI font)
{
	SendMessage(E2W_HWND(win),WM_SETFONT, (WPARAM) E2W_HGDI(font), MAKELPARAM(TRUE,0));
	/* no return value */
}

FRTAPI E_HGDI win_getfont(E_HWND win)
{
	return(W2E_HGDI(SendMessage(E2W_HWND(win),WM_GETFONT,0,0)));
	/* fixme: NULL (sys font) ? */
}

/* background colour */

FRTAPI void win_back_color(E_HWND win, E_HGDI brush)
{
	SetClassLong(E2W_HWND(win),GCL_HBRBACKGROUND,(LONG) E2W_HGDI(brush));
	/* return previous value */
}

/* clipboard */

FRTAPI EBOOL win_clip_open(E_HWND win)
{
	return(W2E_BOOL(OpenClipboard(E2W_HWND(win))));
}

FRTAPI void win_clip_close(void)
{
	BOOL fres;

	fres=CloseClipboard();
	FALSE_EXCEPTION(fres,"win_clip_close");
}

FRTAPI void win_clip_empty(void)
{
	EmptyClipboard();
}

FRTAPI EBOOL win_clip_has(EINT type)
{
	return(W2E_BOOL(IsClipboardFormatAvailable(E2W_INT(type))));
}

/* text clipboard copy/paste */

FRTAPI EBOOL win_clip_hastext(void)
{
	BOOL result;

#ifdef UNICODE
	result = IsClipboardFormatAvailable(CF_UNICODETEXT);
#else
	result = IsClipboardFormatAvailable(CF_TEXT);
#endif
	return(W2E_BOOL(result));
}

FRTAPI void win_clip_copytext(ESTRING text)
{
	HGLOBAL htext;
	HANDLE result;
	LPTSTR etext,htextptr;
	int i;
		
	etext = E2W_STRING(text);

	htext = GlobalAlloc(GMEM_DDESHARE | GMEM_MOVEABLE, (lstrlen(etext)+1) * sizeof(TCHAR));
	NULL_EXCEPTION(htext,"win_clip_copytext.1"); /* fixme: graceful handling? */

	htextptr=(LPTSTR) GlobalLock(htext);
	NULL_EXCEPTION(htextptr,"win_clip_copytext.2");

	/* copy */
	for(i=0;etext[i]!=0;i++)
		htextptr[i]=etext[i];
	htextptr[i]=0;
	
	GlobalUnlock(htext);

#ifdef UNICODE
	result = SetClipboardData(CF_UNICODETEXT,htext);
#else
	result = SetClipboardData(CF_TEXT,htext);
#endif

	/* htext ownership transfered to OS */
	NULL_EXCEPTION(result,"win_clip_copytext.3");
}

FRTAPI void win_clip_pastetext(void)
{
	HANDLE htext;
	LPTSTR stext;
	
	fenestra_void_string();

#ifdef UNICODE
	if((htext=GetClipboardData(CF_UNICODETEXT))!=NULL)
#else
	if((htext=GetClipboardData(CF_TEXT))!=NULL)
#endif
	{
		stext = (LPTSTR) GlobalLock((HGLOBAL) htext);
		NULL_EXCEPTION(stext,"win_clip_pastetext");
		fenestra_copy_string(stext);
		GlobalUnlock((HGLOBAL) htext);
	}
}

/* handle clipboard copy/paste */

FRTAPI void win_clip_copy(EINT type, EINT handle)
{
	HANDLE fres;

	fres=SetClipboardData(E2W_INT(type),(HANDLE) E2W_INT(handle));
	NULL_EXCEPTION(fres,"win_clip_copy");
}

FRTAPI EINT win_clip_paste(EINT type)
{
	return(W2E_INT(GetClipboardData(E2W_INT(type))));
}

/* drag and drop */

FRTAPI void win_drag_accept(E_HWND handle, EBOOL status)
{
	DragAcceptFiles(E2W_HWND(handle),E2W_BOOL(status));
}

FRTAPI EINT win_drag_files(E_HWND drophandle)
{
	return(W2E_INT(DragQueryFile((HDROP) E2W_HWND(drophandle),0xFFFFFFFF,NULL,0)));
}

FRTAPI void win_drag_getfile(E_HWND drophandle, EINT index)
{
	LPTSTR wresult;
	UINT wlen;

	fenestra_void_string();
	
	wlen=DragQueryFile((HDROP) E2W_HWND(drophandle),E2W_INT(index),NULL,0);
	if (wlen>0)
	{
		wresult=fenestra_malloc((wlen+1) * sizeof(TCHAR));

		wlen=DragQueryFile((HDROP) E2W_HWND(drophandle),E2W_INT(index),wresult,wlen+1);
		ASSERT_EXCEPTION(wlen>0, "win_drag_getfile");
		
		fenestra_pass_string(wresult); /* will free */
	}
}

FRTAPI void win_drag_finish(E_HWND drophandle)
{
	DragFinish((HDROP) E2W_HWND(drophandle));
}

FRTAPI EBOOL win_drag_getpoint(E_HWND drophandle)
{
	POINT pt;
	BOOL result;

	result=DragQueryPoint((HDROP) E2W_HWND(drophandle),&pt);
	if(result!=FALSE)
		fenestra_set_int2(pt.x,pt.y);

	return(W2E_BOOL(result));
}

/* common/standard dialogs */

FRTAPI EINT win_message(E_HWND parent, ESTRING title, ESTRING text, EINT icon, EINT buttons)
{
	int result;

	result = MessageBox(E2W_HWND(parent), E2W_STRING(text), E2W_STRING(title), E2W_INT(icon) | E2W_INT(buttons));
	ZERO_EXCEPTION(result,"win_message");

	return(W2E_INT(result));
}

FRTAPI EINT win_std_color(E_HWND parent, EINT oldcolor)
{
	CHOOSECOLOR cc;
	COLORREF colors[16]; /* fixme: save them? */

	cc.lStructSize = sizeof(CHOOSECOLOR);
	cc.hwndOwner = E2W_HWND(parent);
	cc.rgbResult = E2W_INT(oldcolor); /* color */
	cc.lpCustColors = colors;
	cc.Flags = /* CC_PREVENTFULLOPEN | */ CC_RGBINIT | CC_SHOWHELP;

	if(ChooseColor(&cc)==TRUE)
		return(W2E_INT(cc.rgbResult));

	if(CommDlgExtendedError()==0)
		return(W2E_INT(BAD_COLOR));

	WINAPI_EXCEPTION("win_std_color");
	return(W2E_INT(BAD_COLOR));
}

FRTAPI EBOOL win_std_font(E_HWND parent, EINT oldcolor, EBOOL getcolor, EBOOL hasfont, ELOGFONT font)
{
	CHOOSEFONT cf;
	LOGFONT *winfont;
	DWORD flags;
	EBOOL result;
	EINT result1,result2;

	flags = CF_SCREENFONTS | CF_SHOWHELP;
	winfont = E2W_LOGFONT(font);
	
	if(hasfont==ETRUE)
		flags |= CF_INITTOLOGFONTSTRUCT;
	if(getcolor==ETRUE)
		flags |= CF_EFFECTS;


	/* CF_APPLY, CF_FIXEDPITCH, CF_WYSIWYG/SCALABLEONLY/BOTH ?*/

	fenestra_memset(&cf,sizeof(CHOOSEFONT));

	cf.lStructSize = sizeof(CHOOSEFONT);
	cf.hwndOwner = E2W_HWND(parent);
	cf.lpLogFont = winfont;
	cf.hDC = NULL; /*if printer, then set CF_BOTH */
	cf.Flags = flags; 
	cf.rgbColors = E2W_INT(oldcolor);
	cf.nFontType = SCREEN_FONTTYPE; /* return value? change? */

	result=W2E_BOOL(ChooseFont(&cf));

	if(result==ETRUE)
	{
		result1=W2E_INT(cf.rgbColors);
	}
	else if(CommDlgExtendedError()==0)
	{
		result1=W2E_INT(BAD_COLOR);
	}
	else
	{
		WINAPI_EXCEPTION("win_std_font");
	}
	result2=W2E_INT(cf.iPointSize / 10);

	fenestra_set_int2(result1,result2);

	return(result);
}

FRTAPI E_HGDI win_std_print(E_HWND parent, EBOOL getdc, EBOOL getdefault)
{
	PRINTDLG psd;
	
	fenestra_memset(&psd,sizeof(PRINTDLG));
	
	psd.lStructSize = sizeof(PRINTDLG);
	psd.hwndOwner = E2W_HWND(parent);
	psd.hDevMode = NULL;
	psd.hDevNames = NULL;
	
	psd.Flags = PD_PRINTSETUP; /* no range(all/pages/selection), copies, etc */

	if (E2W_BOOL(getdc)==TRUE)
		psd.Flags |= PD_RETURNDC; /* Return a device context */
	else
		psd.Flags |= PD_RETURNIC; /*	Return an information context */
	
	if (E2W_BOOL(getdefault)==TRUE)
		psd.Flags |= PD_RETURNDEFAULT; /* Do not display the dialog and print the device */
	
	if (PrintDlg(&psd)==TRUE)
		return(W2E_HGDI(psd.hDC));
	else
		return(0); /* Invalid handle value, error or user cancel */
}

FRTAPI void win_std_file(EBOOL open,E_HWND parent, ESTRING title, ESTRING initdir, ESTRING initfile, ESTRING filters)
{
	OPENFILENAME of;
	LPTSTR efilters, einitdir, einitfile, etitle;
	LPTSTR filt, result;
	EBOOL resbool;
	int i;

	/* adapt filters */
	efilters = E2W_STRING(filters);
	filt = fenestra_malloc((lstrlen(efilters)+1)*sizeof(TCHAR));
	for(i=0;efilters[i]!=0;i++) {
		if (efilters[i] == OPENFILE_SEPARATOR)
			filt[i] = 0;
		else
			filt[i] = efilters[i];
	}
	filt[i]=0;

	/* empty title is default title */
	etitle = E2W_STRING(title);
	if(*etitle==0)
		etitle = NULL;

	/* empty init dir is current dir */
	einitdir = E2W_STRING(initdir);
	if(*einitdir==0)
		einitdir = NULL;

	/* file */
	result = fenestra_malloc((OPENFILE_FILESIZE+1)*sizeof(TCHAR)); /* Allocate file buffer */
	result[0] = ((TCHAR) 0);

	einitfile = E2W_STRING(initfile);
	for(i=0;(einitfile[i]!=0) && (i<(OPENFILE_FILESIZE));i++)
		result[i]=einitfile[i];
	result[i]=0;

	/* prepare parameters structure */
	fenestra_memset(&of,sizeof(OPENFILENAME));
	
	of.lStructSize = sizeof(OPENFILENAME);
	of.lpstrFilter = filt;
	of.hwndOwner = E2W_HWND(parent);
	of.lpstrFile = result;
	of.nMaxFile = OPENFILE_FILESIZE; /* size of lpstrFile */
	of.lpstrInitialDir = einitdir; 
	of.lpstrTitle = etitle;
	of.Flags = OFN_HIDEREADONLY;

	if (E2W_BOOL(open)==TRUE)
		resbool=GetOpenFileName(&of);
	else
		resbool=GetSaveFileName(&of);

	fenestra_set_bool(resbool);

	fenestra_pass_string(result); /* will free file buffer */
	fenestra_free(filt);
}   


/* button interface */

E_HWND intr_but_create(E_HWND parent, DWORD dstyle) /* fixme: unify? */
{
	HWND ret;
	
	ret=CreateWindow(TEXT("BUTTON"), TEXT("Default"), WS_VISIBLE | WS_CHILD | dstyle,
			0,0,0,0, E2W_HWND(parent), (HMENU)UNIQUEID, PROGHINST, NULL);		
	NULL_EXCEPTION(ret,"intr_but_create");

	return(W2E_HWND(ret));
}

FRTAPI E_HWND win_but_cpush(E_HWND parent)
{
	return(intr_but_create(parent, WS_TABSTOP | BS_PUSHBUTTON));
}

FRTAPI E_HWND win_but_cdefpush(E_HWND parent)
{
	return(intr_but_create(parent, WS_TABSTOP | BS_DEFPUSHBUTTON));
}

FRTAPI E_HWND win_but_cradio(E_HWND parent)
{
	return(intr_but_create(parent, WS_TABSTOP | WS_GROUP | BS_AUTORADIOBUTTON));
}

FRTAPI E_HWND win_but_ccheck(E_HWND parent)
{
	return(intr_but_create(parent, WS_TABSTOP | BS_AUTOCHECKBOX));
}

FRTAPI E_HWND win_but_ctristate(E_HWND parent)
{
	return(intr_but_create(parent, WS_TABSTOP | BS_AUTO3STATE));
}

FRTAPI E_HWND win_but_cgroupbox(E_HWND parent)
{
	return(intr_but_create(parent, BS_GROUPBOX));
}

FRTAPI EBOOL win_but_isdown(E_HWND win)
{
	if(SendMessage(E2W_HWND(win),BM_GETSTATE,0,0) & 0x0004)
		return(ETRUE);
	return(EFALSE);
}

FRTAPI void win_but_default(E_HWND win, EBOOL def)
{
	SendMessage(E2W_HWND(win), BM_SETSTYLE, (WPARAM) (def==ETRUE ? BS_DEFPUSHBUTTON : BS_PUSHBUTTON), MAKELPARAM(TRUE,0)); 
	/* no return value */
}

/* general message processing */

FRTAPI void win_msg_void(E_HWND win, EINT msgid)
{
	SendMessage(E2W_HWND(win),E2W_INT(msgid),0,0);
}

FRTAPI EBOOL win_msg_bool(E_HWND win, EINT msgid)
{
	if(SendMessage(E2W_HWND(win),E2W_INT(msgid),0,0)==0)
		return(EFALSE);
	return(ETRUE);
}

FRTAPI void win_msg_setbool(E_HWND win, EINT msgid, EBOOL stat)
{
	SendMessage(E2W_HWND(win),E2W_INT(msgid),(WPARAM)(stat == ETRUE ? 1 : 0),0);
	/* no return value */
}

FRTAPI EINT win_msg_int(E_HWND win, EINT msgid)
{
	LRESULT result;

	result=SendMessage(E2W_HWND(win),E2W_INT(msgid),0,0); 
	return(W2E_INT(result));	
}

FRTAPI void win_msg_setint(E_HWND win, EINT msgid, EINT value)
{
	SendMessage(E2W_HWND(win),E2W_INT(msgid),E2W_INT(value),0);
}

FRTAPI void win_postmsg_void(E_HWND win, EINT msgid)
{
	BOOL fres;

	fres=PostMessage(E2W_HWND(win),E2W_INT(msgid),0,0);
	FALSE_EXCEPTION(fres, "win_postmsg_void");
}

/* menu */

FRTAPI void win_drawmenu(E_HWND win)
{
	BOOL fres;

	fres=DrawMenuBar(E2W_HWND(win));
	FALSE_EXCEPTION(fres, "win_drawmenu");
}

FRTAPI void win_menu_addsep(E_HMENU menu)
{
	BOOL fres;

	fres=AppendMenu(E2W_HMENU(menu), MF_SEPARATOR, 0, NULL);
	FALSE_EXCEPTION(fres, "win_menu_addsep");
}

FRTAPI void win_menu_additem(E_HMENU menu, EINT id, ESTRING name)
{
	BOOL fres;

	fres=AppendMenu(E2W_HMENU(menu), MF_STRING, (UINT) E2W_INT(id), E2W_STRING(name));
	FALSE_EXCEPTION(fres, "win_menu_additem");
}

FRTAPI void win_menu_addpopup(E_HMENU menu, E_HMENU popup, ESTRING name)
{
	BOOL fres;

	fres=AppendMenu(E2W_HMENU(menu), /*MF_STRING |*/ MF_POPUP, (UINT) E2W_HMENU(popup), E2W_STRING(name));
	FALSE_EXCEPTION(fres, "win_menu_addpopup");
}

FRTAPI E_HMENU win_menu_cbar(void)
{
	HMENU ret;

	ret=CreateMenu();
	NULL_EXCEPTION(ret, "win_menu_cbar");
	
	return(W2E_HMENU(ret));
}

FRTAPI E_HMENU win_menu_cpopup(void)
{
	HMENU ret;
	
	ret=CreatePopupMenu();
	NULL_EXCEPTION(ret,"win_menu_cpopup");

	return(W2E_HMENU(ret));
}

FRTAPI void win_menu_destroy(E_HMENU menu)
{
	BOOL fres;

	fres=DestroyMenu(E2W_HMENU(menu));
	FALSE_EXCEPTION(fres,"win_menu_destroy");
}

FRTAPI void win_menu_delete(E_HMENU menu, EINT id)
{
	BOOL fres;

	fres=DeleteMenu(E2W_HMENU(menu), MF_BYCOMMAND, E2W_INT(id));
	FALSE_EXCEPTION(fres,"win_menu_delete");
}

FRTAPI void win_menu_track(E_HMENU menu, E_HWND parent, EINT x, EINT y)
{
	BOOL fres;
	POINT p;
	
	p.x = (LONG) E2W_INT(x);
	p.y = (LONG) E2W_INT(y);
	
	fres=ClientToScreen(E2W_HWND(parent),&p);
	FALSE_EXCEPTION(fres,"win_menu_track.1");

	fres=TrackPopupMenu(E2W_HMENU(menu),TPM_LEFTALIGN | TPM_LEFTBUTTON,
			(int) p.x,(int) p.y,0,E2W_HWND(parent),NULL);
	FALSE_EXCEPTION(fres,"win_menu_track.2");
}

FRTAPI void win_menu_set(E_HMENU menu, E_HWND parent)
{
	BOOL fres;

	fres=SetMenu(E2W_HWND(parent),E2W_HMENU(menu));
	FALSE_EXCEPTION(fres,"win_menu_set");
}

FRTAPI void win_menu_chtext(E_HMENU menu, EINT id, ESTRING text)
{
	BOOL fres;

	fres=ModifyMenu(E2W_HMENU(menu),E2W_INT(id),MF_BYCOMMAND,
			E2W_INT(id),E2W_STRING(text));
	FALSE_EXCEPTION(fres,"win_menu_chtext");	
}

FRTAPI void win_menu_enable(E_HMENU menu, EINT id, EBOOL val)
{
	EnableMenuItem(E2W_HMENU(menu),(UINT) E2W_INT(id), 
		(val == ETRUE) ? MF_ENABLED : MF_GRAYED);
	/* Documentation on return value inconsistent */
}

FRTAPI void win_menu_checked(E_HMENU menu, EINT id, EBOOL val)
{
	DWORD fres;

	fres=CheckMenuItem(E2W_HMENU(menu),(UINT) E2W_INT(id), 
			(val == ETRUE) ? MF_CHECKED : MF_UNCHECKED);
	ASSERT_EXCEPTION(fres != 0xFFFFFFFF,"win_menu_checked");
}

FRTAPI EBOOL win_menu_isenabled(E_HMENU menu, EINT id)
{
	if(GetMenuState(E2W_HMENU(menu),(UINT) E2W_INT(id),MF_BYCOMMAND) & MF_GRAYED)
		return(EFALSE);
	return(ETRUE);
}

FRTAPI EBOOL win_menu_ischecked(E_HMENU menu, EINT id)
{
	if(GetMenuState(E2W_HMENU(menu),(UINT) E2W_INT(id),MF_BYCOMMAND) & MF_CHECKED)
		return(ETRUE);
	return(EFALSE);
}

/* accelerators */

FRTAPI E_HWND win_accel_create(EPOINTER acc_table, EINT size)
{
	HACCEL result;

	result=CreateAcceleratorTable((LPACCEL) E2W_POINTER(acc_table),E2W_INT(size));
	NULL_EXCEPTION(result,"win_accel_create");

	return(W2E_HWND(result));
}

FRTAPI void win_accel_destroy(E_HWND ahandle)
{
	BOOL result;

	result=DestroyAcceleratorTable((HACCEL) E2W_HWND(ahandle));
	FALSE_EXCEPTION(result,"win_accel_destroy");
}

/* scrollbar (controls) */

FRTAPI E_HWND win_scroll_create(E_HWND parent, EBOOL ishoriz)
{
	HWND ret;

	ret=CreateWindow(TEXT("SCROLLBAR"),TEXT("Default"),
			(ishoriz == ETRUE) ? WS_VISIBLE | WS_CHILD | SBS_TOPALIGN | SBS_HORZ :
				WS_VISIBLE | WS_CHILD | SBS_LEFTALIGN | SBS_VERT,
			0,0,0,0, 
			E2W_HWND(parent),(HMENU)UNIQUEID,PROGHINST,NULL);
	NULL_EXCEPTION(ret,"win_scroll_create");

	return(W2E_HWND(ret));
}

/* scrollbar */

FRTAPI void win_scroll_set(E_HWND win, EINT direction, EINT pos)
{
	int fres;

	fres=SetScrollPos(E2W_HWND(win), E2W_INT(direction), E2W_INT(pos), TRUE);
	ASSERT_EXCEPTION(fres != 0, "win_scroll_set");
}

FRTAPI void win_scroll_range(E_HWND win, EINT direction, EINT range)
{
	BOOL fres;

	fres=SetScrollRange(E2W_HWND(win), E2W_INT(direction), 1,E2W_INT(range), FALSE);
	FALSE_EXCEPTION(fres,"win_scroll_range");
}

FRTAPI EINT win_scroll_get(E_HWND win, EINT direction)
{
	return(W2E_INT( GetScrollPos(E2W_HWND(win),E2W_INT(direction)) ));
}

/* listbox */

FRTAPI E_HWND win_listbox_create(E_HWND parent, EBOOL multi, EBOOL sorted)
{
	HWND ret;
	DWORD style;

	style = WS_VISIBLE | WS_CHILD | LBS_HASSTRINGS | LBS_NOINTEGRALHEIGHT;

	if(sorted == ETRUE)
		style |= LBS_NOTIFY | LBS_SORT | WS_VSCROLL; /* win4: LBS_STANDARD w/o WS_BORDER */
	else
		style |= LBS_NOTIFY | WS_VSCROLL;

	if(multi == ETRUE)
		style |= LBS_EXTENDEDSEL | LBS_MULTIPLESEL;

	ret=CreateWindowEx(WS_EX_CLIENTEDGE,TEXT("LISTBOX"),TEXT("Default"),
				style,
				0,0,0,0, 
				E2W_HWND(parent),(HMENU)UNIQUEID,PROGHINST,NULL);
	NULL_EXCEPTION(ret,"win_listbox_create");

	return(W2E_HWND(ret));
}

FRTAPI void win_listbox_addtext(E_HWND win, ESTRING text)
{
	LRESULT fres;

	fres=SendMessage(E2W_HWND(win),LB_ADDSTRING,0,(LPARAM) E2W_STRING(text));
	ASSERT_EXCEPTION(fres >= 0, "win_listbox_addtext");
}

FRTAPI void win_listbox_selstring(E_HWND win, ESTRING in)
{
	LRESULT fres;

	fres=SendMessage(E2W_HWND(win),LB_SELECTSTRING,(WPARAM) -1, (LPARAM) E2W_STRING(in));
	ASSERT_EXCEPTION(fres != CB_ERR, "win_listbox_selstring");
}


FRTAPI EINT win_listbox_mitems(E_HWND win, EINTARRAY lpdword, EINT lpdword_size)
{
	int *winptr, wsize, i; /* fixme: winptr of a better type? */
	DWORD *out;
	LRESULT result;

	wsize = E2W_INT(lpdword_size);
	out = E2W_INTARRAY(lpdword);

	winptr = fenestra_malloc(sizeof(int) * wsize);
		result = SendMessage(E2W_HWND(win),LB_GETSELITEMS,(WPARAM) wsize,(LPARAM) winptr);
	 	ASSERT_EXCEPTION(result != LB_ERR, "win_listbox_mitems.1");
			
		/* fixme: useful? */
		for (i = 0; (i < ((int)result)) && (i < wsize) ; i++ )
		{
			*(out + i) = *(winptr + i);
		}
		/* fixme: overflow? ASSERT_EXCEPTION(i < wsize,"win_listbox_mitems.2"); */
		
	fenestra_free(winptr);
	
	return(W2E_INT(result));
}

FRTAPI void win_listbox_gettext(E_HWND win, EINT index)
{
	int lenght;
	LPTSTR text;
	LRESULT fres;

	fenestra_void_string();

	lenght=SendMessage(E2W_HWND(win),LB_GETTEXTLEN,(WPARAM) E2W_INT(index),0);
	ASSERT_EXCEPTION(lenght >= 0,"win_listbox_gettext.1");
	
	if(lenght!=0)
	{
		text = fenestra_malloc((lenght + 2) * sizeof(TCHAR));
		text[0]=((TCHAR) 0);
		 
		fres=SendMessage(E2W_HWND(win),LB_GETTEXT,(WPARAM) E2W_INT(index),(LPARAM) text);
		ASSERT_EXCEPTION(fres!=LB_ERR,"win_listbox_gettext.2");
		
		fenestra_pass_string(text); /* will free */
	}
}

/* Edit box */

FRTAPI E_HWND win_editbox_create(E_HWND parent, EBOOL multi, EBOOL wrap)
{
	HWND ret;
	DWORD style;

	style = WS_VISIBLE | WS_CHILD;

	if((multi == EFALSE) && (wrap == EFALSE))
		style |= ES_AUTOHSCROLL;
	else if((multi == ETRUE) && (wrap == EFALSE)) 
		style |= WS_HSCROLL | WS_VSCROLL | ES_MULTILINE | ES_AUTOHSCROLL | ES_AUTOVSCROLL;
	else if((multi == ETRUE) && (wrap == ETRUE))
		style |= WS_VSCROLL | ES_MULTILINE | ES_AUTOVSCROLL;
	else
		WINAPI_EXCEPTION("win_editbox_create");

	ret=CreateWindowEx(WS_EX_CLIENTEDGE,TEXT("EDIT"),TEXT("Default"),
				style,
				0,0,0,0, 
				E2W_HWND(parent),(HMENU)UNIQUEID,PROGHINST,NULL);
	NULL_EXCEPTION(ret,"win_editbox_create");

	return(W2E_HWND(ret));
}

/* Dialog base units */

FRTAPI EINT win_dialog_unitx(void )
{
	return(W2E_INT(LOWORD(GetDialogBaseUnits())));
}

FRTAPI EINT win_dialog_unity(void )
{
	return(W2E_INT(HIWORD(GetDialogBaseUnits())));
}

/* Static controls */

FRTAPI E_HWND win_static_create(E_HWND parent, EINT type)
{
	HWND ret;
	DWORD style;
	
	style = WS_VISIBLE | WS_CHILD | (E2W_INT(type));
	
	ret=CreateWindow(TEXT("STATIC"),TEXT("Default"),
			style,
			0,0,0,0,
			E2W_HWND(parent),(HMENU)UNIQUEID,PROGHINST,NULL);
	NULL_EXCEPTION(ret,"win_static_create");

	return(W2E_HWND(ret));
}

FRTAPI void win_static_icon(E_HWND win, E_HGDI icon)
{
	LRESULT fres;

	fres=SendMessage(E2W_HWND(win),STM_SETICON,(WPARAM) E2W_HGDI(icon),0);
	/*ASSERT_EXCEPTION(fres != 0,"win_static_icon");*/
}

/* Combobox */

FRTAPI E_HWND win_combo_create(E_HWND parent, EBOOL editable, EBOOL sorted)
{
	HWND ret;
	DWORD style;
	
	style = WS_VISIBLE | WS_BORDER | WS_CHILD | CBS_AUTOHSCROLL; /* | CBS_NOINTEGRALHEIGHT */

	if(editable == ETRUE)
		style |= CBS_DROPDOWN;
	else
		style |= CBS_DROPDOWNLIST;
	
	if(sorted == ETRUE)
		style |= CBS_SORT;

	ret=CreateWindow(TEXT("COMBOBOX"),TEXT("Default"),
				style,
				0,0,100/*CW_USEDEFAULT*/,0, 
				E2W_HWND(parent),(HMENU)UNIQUEID,PROGHINST,NULL);
				/* needs random width because sets height itself when width != 0 */
	NULL_EXCEPTION(ret,"win_combo_create");

	return(W2E_HWND(ret));
}

FRTAPI E_HWND win_combo_editctl(E_HWND win)
{
	HWND result;
	POINT pt;

	/* hack to get the edit control from a combobox,
	   this is the documented method from the Win32 SDK,
	   though it recommends using (1,1) which does not work 
	   with 3D controls on the 95 interface... */
	pt.x = 4;
	pt.y = 4;
	result = ChildWindowFromPoint(E2W_HWND(win),pt);
	NULL_EXCEPTION(result,"win_combo_editctl");

	return(W2E_HWND(result));	
}

FRTAPI void win_combo_laddtext(E_HWND win, ESTRING text)
{
	LRESULT fres;

	fres=SendMessage(E2W_HWND(win),CB_ADDSTRING,0,(LPARAM) E2W_STRING(text));
	ASSERT_EXCEPTION(fres >= 0, "win_combo_laddtext");
}

FRTAPI void win_combo_selstring(E_HWND win, ESTRING in)
{
	LRESULT fres;

	fres=SendMessage(E2W_HWND(win),CB_SELECTSTRING,(WPARAM) -1, (LPARAM) E2W_STRING(in));
	ASSERT_EXCEPTION(fres != CB_ERR, "win_combo_selstring");
}

FRTAPI void win_combo_lgettext(E_HWND win, EINT index)
{
	int lenght;
	LRESULT fres;
	LPTSTR text;
	
	fenestra_void_string();

	lenght=SendMessage(E2W_HWND(win),CB_GETLBTEXTLEN,(WPARAM) E2W_INT(index),0);
	ASSERT_EXCEPTION(lenght >= 0,"win_combo_gettext.1");
	
	if(lenght!=0)
	{
		text = fenestra_malloc((lenght + 2) * sizeof(TCHAR));
		text[0]=((TCHAR) 0);

		fres=SendMessage(E2W_HWND(win),CB_GETLBTEXT,(WPARAM) E2W_INT(index),(LPARAM) text);
		ASSERT_EXCEPTION(fres!=LB_ERR,"win_combo_gettext.2");

		fenestra_pass_string(text); /* will free */
	}
}

/* Cursor */

FRTAPI E_HGDI win_cursor_std(EINT code)
{
	HCURSOR ret;

	ret = LoadCursor(NULL,MAKEINTRESOURCE(E2W_INT(code)));
	NULL_EXCEPTION(ret,"win_cursor_std");

	return(W2E_HGDI(ret));
}

FRTAPI E_HGDI win_cursor_rsc(ESTRING name)
{
	HCURSOR ret;

	ret = LoadCursor(PROGHINST,E2W_STRING(name));
	NULL_EXCEPTION(ret,"win_cursor_rsc");

	return(W2E_HGDI(ret));	
}

FRTAPI void win_cursor_set(E_HGDI curs)
{
	SetCursor(E2W_HGDI(curs));
	/* return value irrelevant */	
}

FRTAPI void win_cursor_show(EBOOL show)
{
	ShowCursor(show == ETRUE ? TRUE : FALSE);
	/* return: display counter, -1 for no mouse */
}

/* Icon */

FRTAPI E_HGDI win_icon_rsc(ESTRING name)
{
	HICON ret;

	ret=LoadIcon(PROGHINST,E2W_STRING(name));
	NULL_EXCEPTION(ret,"win_icon_rsc");

	return(W2E_HGDI(ret));
}

FRTAPI void win_icon_set(E_HWND win, E_HGDI icon)
{
	DWORD fres;

	fres=SetClassLong(E2W_HWND(win),GCL_HICON,(LONG) E2W_HGDI(icon));
	ZERO_EXCEPTION(fres,"win_icon_set");	
}

/* Caret */

FRTAPI void win_caret_create(E_HWND win, EINT w, EINT h)
{
	BOOL fres;

	fres=CreateCaret(E2W_HWND(win),NULL,E2W_INT(w),E2W_INT(h));
	FALSE_EXCEPTION(fres,"win_caret_create");
}

FRTAPI void win_caret_destroy(void)
{
	BOOL fres;

	fres=DestroyCaret();
	FALSE_EXCEPTION(fres,"win_caret_destroy");
}

FRTAPI void win_caret_hide(E_HWND win)
{
	BOOL fres;
	
	fres=HideCaret(E2W_HWND(win));
	FALSE_EXCEPTION(fres,"win_caret_hide");
}

FRTAPI void win_caret_show(E_HWND win)
{
	BOOL fres;

	fres=ShowCaret(E2W_HWND(win));
	FALSE_EXCEPTION(fres,"win_caret_show");
}

FRTAPI void win_caret_position(EINT x, EINT y)
{
	BOOL fres;

	fres=SetCaretPos(E2W_INT(x),E2W_INT(y));
	FALSE_EXCEPTION(fres,"win_caret_position");
}

/* Help */

FRTAPI EBOOL win_help_int(E_HWND winid, ESTRING helpfile, EINT command, EINT value)
{
	BOOL result;

	result = WinHelp(E2W_HWND(winid),E2W_STRING(helpfile),(UINT) E2W_INT(command),(DWORD) E2W_INT(value));
	return( result == TRUE ? ETRUE : EFALSE );
}

FRTAPI EBOOL win_help_str(E_HWND winid, ESTRING helpfile, EINT command, ESTRING value)
{
	BOOL result;

	result = WinHelp(E2W_HWND(winid),E2W_STRING(helpfile),(UINT) E2W_INT(command),(DWORD) E2W_STRING(value));
	return( result == TRUE ? ETRUE : EFALSE );
}

/* =================================================================
 * GRAPHICS DEVICE INTERFACE (GDI)
 * =================================================================
 */

/* get a window DC */

FRTAPI E_HGDI win_gdi_get(E_HWND winid)
{
	HDC wdc;

	wdc=GetDC(E2W_HWND(winid));
	NULL_EXCEPTION(wdc,"win_gdi_get");
	return(W2E_HGDI(wdc));
}

/* release a window DC */

FRTAPI void win_gdi_release(E_HGDI dcid, E_HWND winid)
{
	ReleaseDC(E2W_HWND(winid),E2W_HGDI(dcid)); 
	/* result unchecked: normally fails if Class/Window DC */	
} 

/* delete a DC */

FRTAPI void win_gdi_delete(E_HGDI dcid)
{
	BOOL fres;

	fres=DeleteDC(E2W_HGDI(dcid));
	FALSE_EXCEPTION(fres,"win_gdi_delete");
}

/* save/restore a DC */

FRTAPI void win_gdi_save(E_HGDI dcid)
{
	int fres;

	fres = SaveDC(E2W_HGDI(dcid));
	ZERO_EXCEPTION(fres,"win_gdi_save");
}

FRTAPI void win_gdi_restore(E_HGDI dcid)
{
	BOOL fres;

	fres=RestoreDC(E2W_HGDI(dcid),-1);
	FALSE_EXCEPTION(fres,"win_gdi_restore");
}

/* enhmetafile */

FRTAPI E_HGDI win_gdi_emf(ESTRING file, E_HGDI ref_dcid, EINT width, EINT height)
{
	LPTSTR fname;
	RECT r;

	fname = E2W_STRING(file);
	if (*fname == 0)
		fname = NULL;

	SetRect(&r,0,0,E2W_INT(width)+1,E2W_INT(height)+1); 
	/* +1 because windows 'width' != logical width */

	return(W2E_HGDI(CreateEnhMetaFile(E2W_HGDI(ref_dcid),fname,&r,NULL)));
}  

FRTAPI E_HGDI win_gdi_closeemf(E_HGDI dcid)
{
	return(W2E_HGDI(CloseEnhMetaFile(E2W_HGDI(dcid))));
}

FRTAPI void win_gdi_delemf(E_HGDI emfid)
{
	BOOL fres;
	
	fres=DeleteEnhMetaFile(E2W_HGDI(emfid));
	FALSE_EXCEPTION(fres,"win_gdi_closeemf"); /* fixme: over cautious? */
}

FRTAPI E_HGDI win_gdi_getemf(ESTRING name)
{
	return(W2E_HGDI(GetEnhMetaFile(E2W_STRING(name))));
}


FRTAPI void win_gdi_drawemf(E_HGDI dcid, E_HGDI emf, EINT bx, EINT by, EINT bxp, EINT byp)
{
	RECT r;

	SetRect(&r,E2W_INT(bx),E2W_INT(by),E2W_INT(bxp),E2W_INT(byp));
	PlayEnhMetaFile(E2W_HGDI(dcid),E2W_HGDI(emf),&r);
	/* check fine? -- there are reasons for error (W32/W95, ...) */
}

/* bitmap */

FRTAPI E_HGDI win_gdi_bmpdc(E_HGDI dcid)
{
	return(W2E_HGDI(CreateCompatibleDC(E2W_HGDI(dcid))));
}

FRTAPI E_HGDI win_gdi_loadbmp(ESTRING name)
{
	return(W2E_HGDI(LoadBitmap(PROGHINST,E2W_STRING(name))));
}

FRTAPI E_HGDI win_gdi_bmpsize(E_HGDI dcid, EINT w, EINT h)
{
	return(W2E_HGDI(CreateCompatibleBitmap(E2W_HGDI(dcid), E2W_INT(w), E2W_INT(h))));
}

FRTAPI EINT win_gdi_bmpw(E_HGDI dcid, E_HGDI bmpid)
{
	BITMAPINFOHEADER bmp;

	fenestra_memset(&bmp,sizeof(BITMAPINFOHEADER));
	bmp.biSize=sizeof(BITMAPINFOHEADER);
	GetDIBits(E2W_HGDI(dcid),E2W_HGDI(bmpid),0,0,NULL,(BITMAPINFO *)&bmp,DIB_RGB_COLORS);

	return(W2E_INT(bmp.biWidth));	
}

FRTAPI EINT win_gdi_bmph(E_HGDI dcid, E_HGDI bmpid)
{
	BITMAPINFOHEADER bmp;

	fenestra_memset(&bmp,sizeof(BITMAPINFOHEADER));
	bmp.biSize=sizeof(BITMAPINFOHEADER);
	GetDIBits(E2W_HGDI(dcid),E2W_HGDI(bmpid),0,0,NULL,(BITMAPINFO *)&bmp,DIB_RGB_COLORS);

	return(W2E_INT(bmp.biHeight));
}

FRTAPI void win_gdi_bitblt(E_HGDI targetdc, EINT targetx, EINT targety, EINT targetw, EINT targeth, 
					E_HGDI sourceid, EINT sourcex, EINT sourcey)
{	
	BOOL fres;
	  
	fres=BitBlt(E2W_HGDI(targetdc), E2W_INT(targetx), E2W_INT(targety), E2W_INT(targetw), E2W_INT(targeth),
			E2W_HGDI(sourceid), E2W_INT(sourcex), E2W_INT(sourcey),SRCCOPY);
	FALSE_EXCEPTION(fres,"win_gdi_bitblt"); 
}

FRTAPI void win_gdi_strblt(E_HGDI targetdc, EINT targetx, EINT targety, EINT targetw, EINT targeth, 
					E_HGDI sourceid, EINT sourcex, EINT sourcey, EINT sourcew, EINT sourceh)
{																														  
	BOOL fres;

	/* fixme: SetStrechbltMode(E2W_HGDI(targetdc),HALFTONE); then SetBrushOrgEx ?? */
	fres=StretchBlt(E2W_HGDI(targetdc), E2W_INT(targetx), E2W_INT(targety), E2W_INT(targetw), E2W_INT(targeth),
			E2W_HGDI(sourceid), E2W_INT(sourcex), E2W_INT(sourcey), E2W_INT(sourcew), E2W_INT(sourceh), 
			SRCCOPY);
	
	FALSE_EXCEPTION(fres,"win_gdi_strblt"); 
}

/* load .BMP */

FRTAPI E_HGDI win_gdi_bmpload(E_HGDI dcid, ESTRING fname)
{
	HANDLE fhandle;
	HBITMAP result;
	DWORD dw,image_size;

	LPBYTE image;
	BITMAPINFO *bmpinfo;
	BITMAPFILEHEADER fileheader;
	BITMAPINFOHEADER infoheader;

	/* init result */
	result = NULL;

	/* open file */
	fhandle = CreateFile(E2W_STRING(fname),GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_READONLY,NULL);
	if(fhandle!=INVALID_HANDLE_VALUE)
	{
		/* read file header */
		if(ReadFile(fhandle,&fileheader,sizeof(BITMAPFILEHEADER),&dw,NULL)!=FALSE)
		{
			/* read bitmap header */
			if(ReadFile(fhandle,&infoheader,sizeof(BITMAPINFOHEADER),&dw,NULL)!=FALSE)
			{
				/* prepare colours structure */
				bmpinfo = fenestra_malloc(sizeof(BITMAPINFOHEADER)+(sizeof(RGBQUAD)*infoheader.biClrUsed));
				fenestra_memcpy(&(bmpinfo->bmiHeader),&infoheader,sizeof(BITMAPINFOHEADER));
				
				/* load colours */
				if(ReadFile(fhandle,bmpinfo->bmiColors,infoheader.biClrUsed * sizeof(RGBQUAD),&dw,NULL)!=FALSE)
				{
					/* prepare bits */
					image_size = fileheader.bfSize - fileheader.bfOffBits;
					image = fenestra_malloc(image_size);
					
					/* load bits */
					if(ReadFile(fhandle,image,image_size,&dw,NULL)!=FALSE)
					{
						result = CreateDIBitmap(E2W_HGDI(dcid), &infoheader,
							CBM_INIT, image, bmpinfo, DIB_RGB_COLORS);
					}

					/* free bits */
					fenestra_free(image);
				}
				
				/* free info */
				fenestra_free(bmpinfo);
			}
		}
		CloseHandle(fhandle);
	}

	return(W2E_HGDI(result));
}

/* save .BMP */

FRTAPI EBOOL win_gdi_bmpsave(E_HGDI dcid, E_HGDI bmpid, ESTRING fname)
{
	BITMAPINFO *bmpinfo;
	LPBYTE image;
	DWORD bmpinfo_size,image_size,dw;
	HANDLE fhandle;

	BITMAPINFOHEADER infoheader;
	BITMAPFILEHEADER bmpheader;

	EBOOL result;

	result = ETRUE;

	fenestra_memset(&infoheader,sizeof(BITMAPINFOHEADER));
	infoheader.biSize = sizeof(BITMAPINFOHEADER);

	/* Initialise header */
	if(GetDIBits(E2W_HGDI(dcid),E2W_HGDI(bmpid),0,0,NULL,(BITMAPINFO *) &infoheader,DIB_RGB_COLORS)!=0)
	{
		/* fixme: dump new bmpinfo size with palette */
		/* fixme: colorbits = bmpinfo->bmiHeader.biBitCounts; */
		/*if(colorbits<24)
		else bmpinfo_size = sizeof(BITMAPINFOHEADER);*/

		
		/* allocate BITMAPINFOHEADER + palette */
		bmpinfo_size = sizeof(BITMAPINFOHEADER) + (sizeof(RGBQUAD) * infoheader.biClrUsed);
		bmpinfo = fenestra_malloc(bmpinfo_size);
		
		/* copy BMPINFOHEADER to bmpinfo */
		fenestra_memcpy(&(bmpinfo->bmiHeader),&infoheader,sizeof(BITMAPINFOHEADER));

		/* allocate image */
		image_size = bmpinfo->bmiHeader.biSizeImage;
		image = fenestra_malloc(image_size);

		/* retrieve bits and header */
		if(GetDIBits(E2W_HGDI(dcid),E2W_HGDI(bmpid),0,bmpinfo->bmiHeader.biHeight,
			image,bmpinfo,DIB_RGB_COLORS)!=0)
		{
			/* initialise file header */
			fenestra_memset(&bmpheader,sizeof(BITMAPFILEHEADER));
			bmpheader.bfType = 0x4d42; /* "BM" */
			bmpheader.bfSize = sizeof(BITMAPFILEHEADER)+ bmpinfo_size + image_size;
			bmpheader.bfOffBits = sizeof(BITMAPFILEHEADER)+bmpinfo_size;

			fhandle=CreateFile(E2W_STRING(fname),GENERIC_WRITE,0,NULL,
				CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
			
			if (fhandle != INVALID_HANDLE_VALUE)
			{
				if (WriteFile(fhandle,&bmpheader,sizeof(BITMAPFILEHEADER),&dw,NULL)==TRUE)
				{
    				if (WriteFile(fhandle,bmpinfo,bmpinfo_size,&dw,NULL)==TRUE)
    				{
    					if(WriteFile(fhandle,image,image_size,&dw,NULL)==FALSE)
    						result = EFALSE;
    				}	
					else
						result = EFALSE;
				}
				else
					result = EFALSE;

				CloseHandle(fhandle);
			}
			else
				result = EFALSE;
		}
		else
			result = EFALSE;

		/* free image */
		fenestra_free(image);	

		/* free BITMAPINFO */
		fenestra_free(bmpinfo);
	}
	else
		result = EFALSE;

	return(result);
}

/* print */

FRTAPI void win_gdi_startdoc(E_HGDI dcid, ESTRING name)
{
	DOCINFO di;
	int fres;

	fenestra_memset(&di,sizeof(DOCINFO));
	di.cbSize=sizeof(DOCINFO);
	di.lpszDocName=E2W_STRING(name);

	fres=StartDoc(E2W_HGDI(dcid),&di);
	
	ASSERT_EXCEPTION(fres > 0,"win_gdi_startdoc"); /* fixme: manage errors? */
}

FRTAPI void win_gdi_enddoc(E_HGDI dcid)
{
	int fres;

	fres=EndDoc(E2W_HGDI(dcid));
	ASSERT_EXCEPTION(fres > 0,"win_gdi_enddoc");
	
}

FRTAPI void win_gdi_startpage(E_HGDI dcid)
{
	int fres;

	fres=StartPage(E2W_HGDI(dcid));
	ASSERT_EXCEPTION(fres > 0,"win_gdi_startpage");
}

FRTAPI void win_gdi_endpage(E_HGDI dcid)
{
	int fres;

	fres=EndPage(E2W_HGDI(dcid));
	ASSERT_EXCEPTION(fres > 0,"win_gdi_endpage");
}

/* beginpaint - endpaint */

FRTAPI E_HGDI win_beginpaint(E_HWND win, EPOINTER ps)
{
	HDC dc;

	dc = BeginPaint(E2W_HWND(win),(LPPAINTSTRUCT) E2W_POINTER(ps));
	NULL_EXCEPTION(dc,"win_beginpaint");

	return(W2E_HGDI(dc));
}

/* endpaint */

FRTAPI void win_endpaint(E_HWND win, EPOINTER ps)
{
	EndPaint(E2W_HWND(win),(LPPAINTSTRUCT) E2W_POINTER(ps));
	/* no return value */
}

/* select object */

FRTAPI void win_gdi_select(E_HGDI dcid, E_HGDI objid)
{
	SelectObject(E2W_HGDI(dcid), E2W_HGDI(objid));
	/* result unckecked: result may vary according to actual type(!) */
}

/* delete object */

FRTAPI void win_gdi_remove(E_HGDI gdiobj)
{
	BOOL fres;
	
	fres = DeleteObject(E2W_HGDI(gdiobj));
	FALSE_EXCEPTION(fres,"win_gdi_remove");
}

/* stock object */

FRTAPI E_HGDI win_gdi_stock(EINT code)
{
	HGDIOBJ result;
	
	result = GetStockObject(E2W_INT(code));
	NULL_EXCEPTION(result,"win_gdi_stock");

	return(W2E_HGDI(result));
}

/* font */

FRTAPI E_HGDI win_gdi_font(ELOGFONT ft)
{
	HFONT result;

	result = CreateFontIndirect(E2W_LOGFONT(ft));
	NULL_EXCEPTION(result,"win_gdi_font");

	return(W2E_HGDI(result));	
}

/* text */

FRTAPI EINT win_gdi_textmetr(E_HGDI dcid, EINT value)
{
	TEXTMETRIC tm;
	LONG result;
	BOOL fres;

	fres=GetTextMetrics(E2W_HGDI(dcid), &tm);
	FALSE_EXCEPTION(fres,"win_gdi_txtmetr.1");

	switch (E2W_INT(value))
	{
		case 1: /* Ftm_overhang */		result = tm.tmOverhang; 			break;
		case 2: /* Ftm_breakchar */		result = (LONG) tm.tmBreakChar;		break;
		case 3:	/* Ftm_defaultchar */	result = (LONG) tm.tmDefaultChar;	break;		
		case 4: /* Ftm_firstchar */		result = (LONG) tm.tmFirstChar;		break;
		case 5: /* Ftm_lastchar */		result = (LONG)tm.tmLastChar;		break;
		case 6:	/* Ftm_avecharw */		result = tm.tmAveCharWidth;			break;
		case 7:	/* Ftm_maxcharw */		result = tm.tmMaxCharWidth;			break;
		case 8: /* Ftm_intleading */	result = tm.tmInternalLeading;		break;
		case 9: /* Ftm_extleading */	result = tm.tmExternalLeading;		break;
		case 10: /* Ftm_height */		result = tm.tmHeight;				break;
		case 11: /* Ftm_weight */		result = tm.tmWeight;				break;
		case 12: /* Ftm_ascent */		result = tm.tmAscent;				break;
		case 13: /* Ftm_descent */		result = tm.tmDescent;				break;
		case 14: /* Ftm_digaspx */		result = tm.tmDigitizedAspectX;		break;
		case 15: /* Ftm_digaspy */		result = tm.tmDigitizedAspectY;		break;
		case 16: /* Ftm_italic */		result = (LONG) tm.tmItalic;		break;
		case 17: /* Ftm_underline*/		result = (LONG) tm.tmUnderlined;	break;
		case 18: /* Ftm_strikeout */	result = (LONG) tm.tmStruckOut;	  	break;
		case 19: /* Ftm_pitch */		result = (LONG) (tm.tmPitchAndFamily & 0x0F); break;
		case 20: /* Ftm_family */		result = (LONG) (tm.tmPitchAndFamily & 0xF0); break;
		case 21: /* Ftm_charset */		result = (LONG) tm.tmCharSet;		break;
		case 22: /* Ftm_fixedpitch */	result = (LONG) (tm.tmPitchAndFamily & TMPF_FIXED_PITCH); break;
		case 23: /* Ftm_vector */		result = (LONG) (tm.tmPitchAndFamily & TMPF_VECTOR); break;
		case 24: /* Ftm_truetype */		result = (LONG) (tm.tmPitchAndFamily & TMPF_TRUETYPE); break;
		case 25: /* Ftm_device */		result = (LONG) (tm.tmPitchAndFamily & TMPF_DEVICE); break;

		default:
		WINAPI_EXCEPTION("win_gdi_txtmetr.2");	
	}
	return(W2E_INT(result));
}

FRTAPI void win_gdi_textface(E_HGDI dcid)
{
	TCHAR fname[LF_FACESIZE+2];
	int fres;

	fres=GetTextFace(E2W_HGDI(dcid),LF_FACESIZE,fname);
	ZERO_EXCEPTION(fres,"win_gdi_textface");
	
	fenestra_copy_string(fname);
}

FRTAPI void win_gdi_text(E_HGDI dcid, EINT x, EINT y, ESTRING text)
{
 	BOOL fres;
 	LPTSTR s = E2W_STRING(text);

	fres = TextOut( E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y), s, lstrlen(s));
	FALSE_EXCEPTION(fres,"win_gdi_text");	
}

FRTAPI void win_gdi_extext(E_HGDI dcid, EINT px, EINT py, EINT rx, EINT ry, EINT rxp, EINT ryp, EBOOL clip, EBOOL fill, ESTRING text)
{
	BOOL fres;
	UINT options = 0;
	LPTSTR s = E2W_STRING(text);
	RECT r;

	if(clip == ETRUE) options |= ETO_CLIPPED;
	if(fill == ETRUE) options |= ETO_OPAQUE;
	SetRect(&r,E2W_INT(rx),E2W_INT(ry),E2W_INT(rxp),E2W_INT(ryp));

	fres = ExtTextOut( E2W_HGDI(dcid), E2W_INT(px), E2W_INT(py), options, &r, 
		s, lstrlen(s), NULL);
	FALSE_EXCEPTION(fres,"win_gid_extext");
}

FRTAPI void win_gdi_textis(E_HGDI dcid, EINT ispace)
{
	SetTextCharacterExtra( E2W_HGDI(dcid), E2W_INT(ispace) );
	/* irrelevant return value */
}

FRTAPI void win_gdi_textclr(E_HGDI dcid, EINT color)
{
	COLORREF fres;
	
	fres=SetTextColor( E2W_HGDI(dcid), E2W_INT(color) );
	ASSERT_EXCEPTION(fres!=CLR_INVALID, "win_gdi_textclr");
}

FRTAPI void win_gdi_textalign(E_HGDI dcid, EINT horiz, EINT vert)
{
	UINT fres;

	fres=SetTextAlign( E2W_HGDI(dcid), E2W_INT(horiz) | E2W_INT(vert));
	ASSERT_EXCEPTION(fres!=GDI_ERROR, "win_gdi_textalign");
}

FRTAPI void win_gdi_textjustif(E_HGDI dcid, EINT justif, EINT brkchar)
{
	BOOL fres;

	fres = SetTextJustification(E2W_HGDI(dcid),E2W_INT(justif), E2W_INT(brkchar));
	FALSE_EXCEPTION(fres,"win_gdi_textjustif");
}

/* get text size, returns 2 static ints */

FRTAPI void win_gdi_textsize(E_HGDI dcid, ESTRING text)
{
	LPTSTR s = E2W_STRING(text);
	SIZE tsize;
	BOOL fres;

	fres=GetTextExtentPoint32( E2W_HGDI(dcid), s, lstrlen(s), &tsize);
	FALSE_EXCEPTION(fres,"win_gdi_textsize");

	fenestra_set_int2(W2E_INT(tsize.cx), W2E_INT(tsize.cy));
}

FRTAPI EINT win_gdi_charsize(E_HGDI dcid, EINT charcode)
{
	UINT code = E2W_INT(charcode);
	INT result;
	BOOL fres;

	fres=GetCharWidth32( E2W_HGDI(dcid), code, code, &result);
	FALSE_EXCEPTION(fres,"win_gdi_charsize");

	return(W2E_INT(result));
}

#ifdef UNICODE
#error "obsolete in full unicode"
#else

/* (Unicode) print text */

FRTAPI void win_gdi_wtext(EPOINTER charmap_ptr, E_HGDI dcid, EINT x, EINT y, EPOINTER text_ptr)
{
	BOOL fres;
	DWORD *charmap;
	DWORD cnt,i;
	LPTSTR cstr;
	LPWSTR wstr;
	
	charmap = E2W_INTARRAY(charmap_ptr);
	cstr = E2W_STRING(text_ptr);
	cnt = lstrlen(cstr);
	
	wstr = fenestra_malloc(sizeof(WCHAR)*(cnt+1));

	cnt = lstrlen(cstr);
	for(i=0;i<cnt;i++) {
		wstr[i]=(WCHAR) charmap[((BYTE) cstr[i])];
	}
	wstr[cnt]=(WCHAR) 0;
	
	fres=TextOutW( E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y), wstr, cnt);
	FALSE_EXCEPTION(fres, "win_gdi_wtext");	

	fenestra_free(wstr);
}

/* (Unicode) extended print text */

FRTAPI void win_gdi_wextext(EPOINTER charmap_ptr, E_HGDI dcid, EINT px, EINT py, EINT rx, EINT ry, EINT rxp, EINT ryp, EBOOL clip, EBOOL fill, ESTRING text_ptr)
{
	BOOL fres;
	UINT options = 0;
	RECT r;
	DWORD *charmap;
	DWORD cnt,i;
	LPTSTR cstr;
	LPWSTR wstr;
	
	charmap = E2W_INTARRAY(charmap_ptr);
	cstr = E2W_STRING(text_ptr);
	cnt = lstrlen(cstr);
	
	wstr = fenestra_malloc(sizeof(WCHAR)*(cnt+1));
	for(i=0;i<cnt;i++) {
		wstr[i]=(WCHAR) charmap[((BYTE) cstr[i])];
	}
	wstr[cnt]=(WCHAR) 0;

	if(clip == ETRUE) options |= ETO_CLIPPED;
	if(fill == ETRUE) options |= ETO_OPAQUE;
	SetRect(&r,E2W_INT(rx),E2W_INT(ry),E2W_INT(rxp),E2W_INT(ryp));

	fres = ExtTextOutW( E2W_HGDI(dcid), E2W_INT(px), E2W_INT(py), options, &r, 
		wstr, cnt, NULL);
	FALSE_EXCEPTION(fres,"win_gid_extext");

	fenestra_free(wstr);
}


/* (Unicode) get text size, returns 2 static ints */

FRTAPI void win_gdi_wtextsize(EPOINTER charmap_ptr, E_HGDI dcid, EPOINTER text_ptr)
{
	BOOL fres;
	DWORD *charmap;
	DWORD cnt,i;
	LPTSTR cstr;
	LPWSTR wstr;
	SIZE tsize;

	charmap = E2W_INTARRAY(charmap_ptr);
	cstr = E2W_STRING(text_ptr);
	cnt = lstrlen(cstr);
	
	wstr = fenestra_malloc(sizeof(WCHAR)*(cnt+1));
	for(i=0;i<cnt;i++) {
		wstr[i]=(WCHAR) charmap[((BYTE) cstr[i])];
	}
	wstr[cnt]=(WCHAR) 0;

	fres=GetTextExtentPoint32W( E2W_HGDI(dcid), wstr, cnt, &tsize);
	FALSE_EXCEPTION(fres, "win_gdi_wtextsize");

	fenestra_free(wstr);
	fenestra_set_int2(W2E_INT(tsize.cx), W2E_INT(tsize.cy));
}

#endif /* !UNICODE */


/* pen */

FRTAPI E_HGDI win_gdi_expen(EINT type, EINT width, EINT style, EINT endstyle, EINT joinstyle,
				EINT colour,
				EINT brushstyle, EINT brushpat)
{
	LOGBRUSH br;
	HPEN pen;

	br.lbStyle = E2W_INT(brushstyle);
	br.lbColor = E2W_INT(colour);
	br.lbHatch = E2W_INT(brushpat);
	
	pen=ExtCreatePen(E2W_INT(type) | E2W_INT(style) | E2W_INT(endstyle) | E2W_INT(joinstyle),
					(DWORD) E2W_INT(width),&br,0,NULL);
	NULL_EXCEPTION(pen,"win_gdi_expen");
	
	return(W2E_HGDI(pen));
}

FRTAPI E_HGDI win_gdi_pen(EINT style, EINT width, EINT colour)
{
	HPEN pen;

	pen=CreatePen(E2W_INT(style),E2W_INT(width),E2W_INT(colour));
	NULL_EXCEPTION(pen,"win_gdi_pen");

	return(W2E_HGDI(pen));
}

/* brush */

FRTAPI E_HGDI win_gdi_brush(EINT brushstyle, EINT brushpat, EINT colour)
{
	LOGBRUSH br;
	HBRUSH brush;

	br.lbStyle = E2W_INT(brushstyle);
	br.lbColor = E2W_INT(colour);
	br.lbHatch = E2W_INT(brushpat);

	brush=CreateBrushIndirect(&br);
	NULL_EXCEPTION(brush,"win_gdi_brush");
	return(W2E_HGDI(brush));
}

/* color */

FRTAPI E_HGDI win_gdi_palette(ELOGPALETTE ptr)
{
	HPALETTE result;

	result = CreatePalette(E2W_LOGPALETTE(ptr));
	NULL_EXCEPTION(result, "win_gdi_palette");

	return(W2E_HGDI(result));
}

FRTAPI void win_gdi_selectpal(E_HGDI dcid, E_HGDI palid)
{
	HPALETTE fres;
	UINT fres1;
	
	fres=SelectPalette(E2W_HGDI(dcid),E2W_HGDI(palid),FALSE);
	NULL_EXCEPTION(fres,"win_gdi_selectpal.1");

	fres1=RealizePalette(E2W_HGDI(dcid));
	ASSERT_EXCEPTION(fres1 != GDI_ERROR, "win_gdi_selectpal.2");
}

FRTAPI EINT win_gdi_rgbcolor(EINT red, EINT green, EINT blue)
{
	return(W2E_INT(RGB((BYTE) E2W_INT(red),(BYTE) E2W_INT(green),(BYTE) E2W_INT(blue))));
}

FRTAPI EINT win_gdi_rgbpalette(EINT red, EINT green, EINT blue)
{
	return(W2E_INT(PALETTERGB((BYTE) E2W_INT(red),(BYTE) E2W_INT(green),(BYTE) E2W_INT(blue))));
}

FRTAPI EINT win_gdi_palcolor(EINT index)
{
	return(W2E_INT(PALETTEINDEX((WORD) E2W_INT(index))));
}

/* mapping */

FRTAPI void win_gdi_mapmode(E_HGDI dcid, EINT mode)
{
	int fres;
	
	fres=SetMapMode(E2W_HGDI(dcid), E2W_INT(mode));
	ZERO_EXCEPTION(fres,"win_gdi_mapmode");
}

FRTAPI void win_gdi_origin(E_HGDI dcid, EINT x, EINT y)
{
	BOOL fres;

	fres=SetWindowOrgEx( E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y), NULL);
	FALSE_EXCEPTION(fres,"win_gdi_origin");
}

FRTAPI void win_gdi_winext(E_HGDI dcid, EINT x, EINT y)
{
	BOOL fres;

	fres=SetWindowExtEx(E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y), NULL);
	FALSE_EXCEPTION(fres,"win_gdi_winext");
}

FRTAPI void win_gdi_vportext(E_HGDI dcid, EINT x, EINT y)
{
	BOOL fres;

	fres=SetViewportExtEx(E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y), NULL);
	FALSE_EXCEPTION(fres,"win_gdi_vportext");
}

FRTAPI void win_gdi_dp2lp(E_HGDI dcid, EINT x, EINT y)
{
	POINT cpoint;
	BOOL fres;
	
	cpoint.x = E2W_INT(x);
	cpoint.y = E2W_INT(y);
	
	fres=DPtoLP(E2W_HGDI(dcid),&cpoint,1);
	FALSE_EXCEPTION(fres,"win_gdi_dp2lp");

	fenestra_set_int2(W2E_INT(cpoint.x), W2E_INT(cpoint.y));
}

FRTAPI void win_gdi_lp2dp(E_HGDI dcid, EINT x, EINT y)
{
	POINT cpoint;
	BOOL fres;

	cpoint.x = E2W_INT(x);
	cpoint.y = E2W_INT(y);

	fres=LPtoDP(E2W_HGDI(dcid),&cpoint,1);
	FALSE_EXCEPTION(fres,"win_gdi_lp2dp");

	fenestra_set_int2(W2E_INT(cpoint.x), W2E_INT(cpoint.y));
}

/**/

FRTAPI void win_gdi_rop2(E_HGDI dcid, EINT mode)
{
	int fres;

	fres=SetROP2(E2W_HGDI(dcid),E2W_INT(mode));
	ZERO_EXCEPTION(fres,"win_gdi_rop2");
}

FRTAPI void win_gdi_bkmode(E_HGDI dcid, EINT mode)
{
	int fres;

	fres=SetBkMode(E2W_HGDI(dcid),E2W_INT(mode));
	ZERO_EXCEPTION(fres,"win_gdi_bkmode");
}

FRTAPI void win_gdi_bkcolor(E_HGDI dcid, EINT color)
{
	COLORREF fres;

	fres=SetBkColor(E2W_HGDI(dcid),E2W_INT(color));
	ASSERT_EXCEPTION(fres!=CLR_INVALID,"win_gdi_bkcolor");
}

/* draw */

FRTAPI void win_gdi_line(E_HGDI dcid, EINT x, EINT y, EINT xp, EINT yp)
{
	BOOL fres;

	fres=MoveToEx(E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y), NULL);
	FALSE_EXCEPTION(fres,"win_gdi_line.1");
	fres=LineTo(E2W_HGDI(dcid), E2W_INT(xp), E2W_INT(yp));
	FALSE_EXCEPTION(fres,"win_gdi_line.2");
} 

FRTAPI void win_gdi_polyline(E_HGDI dcid, EINTARRAY lpdword_points, EINT lpdword_points_size)
{
	int i,wsize;
	POINT *wpoints;
	POINT *cp;
	BOOL fres;

	wsize = E2W_INT(lpdword_points_size) / 2;
	wpoints = fenestra_malloc(sizeof(POINT) * wsize);

		for (i=0; i<wsize; i++)
		{
			cp = wpoints + i;
			cp->x = *(E2W_INTARRAY(lpdword_points) + (i * 2));
			cp->y = *(E2W_INTARRAY(lpdword_points) + (i * 2) + 1);
		}

		fres=Polyline(E2W_HGDI(dcid), wpoints, wsize);
		FALSE_EXCEPTION(fres,"win_gdi_polyline");

	fenestra_free(wpoints);
}

FRTAPI void win_gdi_bezier(E_HGDI dcid, EINTARRAY lpdword_points, EINT lpdword_points_size)
{
	int i,wsize;
	POINT *wpoints;
	POINT *cp;
	BOOL fres;

	wsize = E2W_INT(lpdword_points_size) / 2;
	wpoints = fenestra_malloc(sizeof(POINT) * wsize);

		for (i=0; i<wsize; i++)
		{
			cp = wpoints + i;
			cp->x = *(E2W_INTARRAY(lpdword_points) + (i * 2));
			cp->y = *(E2W_INTARRAY(lpdword_points) + (i * 2) + 1);
		}

		fres=PolyBezier(E2W_HGDI(dcid), wpoints, wsize);
		FALSE_EXCEPTION(fres,"win_gdi_bezier");

	fenestra_free(wpoints);
}

FRTAPI void win_gdi_polygon(E_HGDI dcid, EINTARRAY lpdword_points, EINT lpdword_points_size)
{
	int i,wsize;
	POINT *wpoints;
	POINT *cp;
	BOOL fres;

	wsize = E2W_INT(lpdword_points_size) / 2;
	wpoints = fenestra_malloc(sizeof(POINT) * wsize);

		for (i=0; i<wsize; i++)
		{
			cp = wpoints + i;
			cp->x = *(E2W_INTARRAY(lpdword_points) + (i * 2));
			cp->y = *(E2W_INTARRAY(lpdword_points) + (i * 2) + 1);
		}

		fres=Polygon(E2W_HGDI(dcid), wpoints, wsize);
		FALSE_EXCEPTION(fres,"win_gdi_polygon");

	fenestra_free(wpoints);
}

FRTAPI void win_gdi_polymode(E_HGDI dcid, EBOOL alternate)
{
	int fres;

	fres=SetPolyFillMode(E2W_HGDI(dcid), (alternate == ETRUE) ? ALTERNATE : WINDING );
	ZERO_EXCEPTION(fres,"win_gdi_polymode");
}

FRTAPI void win_gdi_focusrect(E_HGDI dcid, EINT rx, EINT ry, EINT rxp, EINT ryp)
{
	RECT r;
	BOOL fres;

	SetRect(&r,E2W_INT(rx),E2W_INT(ry),E2W_INT(rxp),E2W_INT(ryp));
	fres=DrawFocusRect(E2W_HGDI(dcid),&r);
	FALSE_EXCEPTION(fres,"win_gdi_focusrect");	
}

FRTAPI void win_gdi_invertrect(E_HGDI dcid, EINT rx, EINT ry, EINT rxp, EINT ryp)
{
	RECT r;
	BOOL fres;

	SetRect(&r, E2W_INT(rx),E2W_INT(ry),E2W_INT(rxp),E2W_INT(ryp));
	fres=InvertRect(E2W_HGDI(dcid),&r);
	FALSE_EXCEPTION(fres,"win_gdi_invertrect");	
}

FRTAPI void win_gdi_rectangle(E_HGDI dcid, EINT rx1, EINT ry1, EINT rx2, EINT ry2)
{
	BOOL fres;

	fres=Rectangle(E2W_HGDI(dcid),E2W_INT(rx1),E2W_INT(ry1),E2W_INT(rx2),E2W_INT(ry2));
	FALSE_EXCEPTION(fres,"win_gdi_rectangle");
} 

FRTAPI void win_gdi_roundrect(E_HGDI dcid,EINT rx1, EINT ry1, EINT rx2, EINT ry2, EINT cw, EINT ch)
{
	BOOL fres;

	fres=RoundRect(E2W_HGDI(dcid),E2W_INT(rx1),E2W_INT(ry1),E2W_INT(rx2),E2W_INT(ry2),
				E2W_INT(cw), E2W_INT(ch));
	FALSE_EXCEPTION(fres,"win_gdi_roundrect");
} 

FRTAPI void win_gdi_ellipse(E_HGDI dcid, EINT rx1, EINT ry1, EINT rx2, EINT ry2)
{
	BOOL fres;

	fres=Ellipse(E2W_HGDI(dcid),E2W_INT(rx1),E2W_INT(ry1),E2W_INT(rx2),E2W_INT(ry2));
	FALSE_EXCEPTION(fres,"win_gdi_ellipse");
}

FRTAPI void win_gdi_arc(E_HGDI dcid, EINT rx1, EINT ry1, EINT rx2, EINT ry2,
					EINT sx, EINT sy, EINT fx, EINT fy)
{
	BOOL fres;

	fres=Arc(E2W_HGDI(dcid),E2W_INT(rx1),E2W_INT(ry1),E2W_INT(rx2),E2W_INT(ry2),
			E2W_INT(sx), E2W_INT(sy), E2W_INT(fx), E2W_INT(fy) );
	FALSE_EXCEPTION(fres,"win_gdi_arc");
}

FRTAPI void win_gdi_chord(E_HGDI dcid, EINT rx1, EINT ry1, EINT rx2, EINT ry2,
					EINT sx, EINT sy, EINT fx, EINT fy)
{
	BOOL fres;

	fres=Chord(E2W_HGDI(dcid),E2W_INT(rx1),E2W_INT(ry1),E2W_INT(rx2),E2W_INT(ry2),
			E2W_INT(sx), E2W_INT(sy), E2W_INT(fx), E2W_INT(fy) );
	FALSE_EXCEPTION(fres,"win_gdi_chord");
}

FRTAPI void win_gdi_pie(E_HGDI dcid, EINT rx1, EINT ry1, EINT rx2, EINT ry2,
					EINT sx, EINT sy, EINT fx, EINT fy)
{
	BOOL fres;

	fres=Pie(E2W_HGDI(dcid),E2W_INT(rx1),E2W_INT(ry1),E2W_INT(rx2),E2W_INT(ry2),
			E2W_INT(sx), E2W_INT(sy), E2W_INT(fx), E2W_INT(fy) );
	FALSE_EXCEPTION(fres,"win_gdi_pie");
}

FRTAPI void win_gdi_arcdir(E_HGDI dcid, EBOOL clockw)
{
	int fres;

	fres=SetArcDirection(E2W_HGDI(dcid), clockw == ETRUE ? AD_CLOCKWISE : AD_COUNTERCLOCKWISE);
	ZERO_EXCEPTION(fres,"win_gdi_arcdir");
}

FRTAPI void win_gdi_icon(E_HGDI dcid, E_HGDI iconid, EINT x, EINT y)
{
	BOOL fres;

	fres=DrawIcon(E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y), E2W_HGDI(iconid));
	FALSE_EXCEPTION(fres,"win_gdi_icon");
}

FRTAPI EBOOL win_gdi_floodfill(E_HGDI dcid, EINT x, EINT y, EINT color, EINT mode)
{
	return(W2E_BOOL(ExtFloodFill(E2W_HGDI(dcid),E2W_INT(x),E2W_INT(y),
					E2W_INT(color),E2W_INT(mode))));
}

/* pixels */

FRTAPI EINT win_gdi_getpixel(E_HGDI dcid, EINT x, EINT y)
{
	int result;
	
	result=GetPixel(E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y));
	if(result == CLR_INVALID)
		return(W2E_INT(BAD_COLOR));

	return(W2E_INT(result));
}

FRTAPI void win_gdi_setpixel(E_HGDI dcid, EINT x, EINT y, EINT color)
{
	BOOL fres;

	fres=SetPixelV(E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y), E2W_INT(color));
	/* win95/nt4: return FALSE? -- FALSE_EXCEPTION(fres,"win_gdi_setpixel"); */
}

/* clipping */

FRTAPI EBOOL win_gdi_ptvisu(E_HGDI dcid, EINT x, EINT y)
{
	return(W2E_BOOL(PtVisible(E2W_HGDI(dcid), E2W_INT(x), E2W_INT(y))));
}

FRTAPI EBOOL win_gdi_rectvisu(E_HGDI dcid, EINT rx1, EINT ry1, EINT rx2, EINT ry2)
{
	RECT r;

	SetRect(&r,E2W_INT(rx1),E2W_INT(ry1),E2W_INT(rx2),E2W_INT(ry2));
	return(W2E_BOOL(RectVisible(E2W_HGDI(dcid),&r)));
}

/* device capabilities */

FRTAPI EINT win_gdi_capint(E_HGDI dcid, EINT cap)
{
	return(W2E_INT(GetDeviceCaps(E2W_HGDI(dcid), E2W_INT(cap))));
}

FRTAPI EBOOL win_gdi_capbool(E_HGDI dcid, EINT cap, EINT field)
{
	EBOOL result = EFALSE;
	
	if( GetDeviceCaps(E2W_HGDI(dcid), E2W_INT(cap)) & field )
		result = ETRUE;

	return(result);
}

/* =================================================================
 * SYSTEM SERVICES
 * =================================================================
 */

/* bit operations */

FRTAPI EINT win_bitand(EINT a, EINT b)
{
	return(W2E_INT( E2W_INT(a) & E2W_INT(b)));
}

FRTAPI EINT win_bitor(EINT a, EINT b)
{
	return(W2E_INT( E2W_INT(a) | E2W_INT(b)));
}

FRTAPI EINT win_bitxor(EINT a, EINT b)
{
	return(W2E_INT( E2W_INT(a) ^ E2W_INT(b)));
}

/* memory block operations */

FRTAPI EINT win_getbyte(EPOINTER ptr, EINT offset)
{
	BYTE *bptr;

	bptr = (BYTE *) E2W_POINTER(ptr);
	bptr += E2W_INT(offset);
	return(W2E_INT(((unsigned int) *bptr)));
}

FRTAPI void win_setbyte(EPOINTER buf, EINT offset, EINT value)
{
	BYTE *bptr;

	bptr = (BYTE *) E2W_POINTER(buf);
	bptr += E2W_INT(offset);
	*bptr = (BYTE) E2W_INT(value);
}

FRTAPI EINT win_getword(EBYTEARRAY ptr)
{
	WORD *w;
	
	w=(WORD *) E2W_BYTEARRAY(ptr);
	return(W2E_INT(*w));
}

FRTAPI EINT win_getlong(EBYTEARRAY ptr)
{
	DWORD *dw;

	dw=(DWORD *) E2W_BYTEARRAY(ptr);
	return(W2E_INT(*dw));
}

FRTAPI void win_setword(EBYTEARRAY buf, EINT index, EINT value)
{
	BYTE *ptr;
	WORD *wo;
	
	ptr=E2W_BYTEARRAY(buf);
	ptr+=E2W_INT(index);
	wo=(WORD *)ptr;
	*wo=(WORD) E2W_INT(value);
}

FRTAPI void win_setlong(EBYTEARRAY buf, EINT index, EINT value)
{
	BYTE *ptr;
	DWORD *dw;
	
	ptr=E2W_BYTEARRAY(buf);
	ptr+=E2W_INT(index);
	dw=(DWORD *)ptr;
	*dw=(DWORD) E2W_INT(value);
}

#define GET_BYTE(val,bit) ((BYTE) ((val >> bit) & 0xFF));

FRTAPI void win_setlong_l(EBYTEARRAY buf, EINT index, EINT value)
{
	BYTE *ptr;
	DWORD dw;

	ptr=E2W_BYTEARRAY(buf);
	ptr+=E2W_INT(index);
	dw=E2W_INT(value);
	
	*(ptr++)=GET_BYTE(dw,0); /* little indian */
	*(ptr++)=GET_BYTE(dw,8);
	*(ptr++)=GET_BYTE(dw,16);
	*(ptr++)=GET_BYTE(dw,24);
}

FRTAPI void win_setlong_b(EBYTEARRAY buf, EINT index, EINT value)
{
	BYTE *ptr;
	DWORD dw;

	ptr=E2W_BYTEARRAY(buf);
	ptr+=E2W_INT(index);
	dw=E2W_INT(value);
	
	*(ptr++)=GET_BYTE(dw,24); /* big indian */
	*(ptr++)=GET_BYTE(dw,16);
	*(ptr++)=GET_BYTE(dw,8);
	*(ptr++)=GET_BYTE(dw,0);
}

/* muldiv */

FRTAPI EINT win_muldiv(EINT m1, EINT m2, EINT d1)
{
	return(W2E_INT(MulDiv(E2W_INT(m1),E2W_INT(m2),E2W_INT(d1))));
}

/* version */

FRTAPI EBOOL win_ver_above(EINT major, EINT minor)
{
	OSVERSIONINFO vi;
	BOOL fres;
	EBOOL result;
	DWORD ma,mi;

	result = EFALSE;
	ma = E2W_INT(major);
	mi = E2W_INT(minor);

	vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	fres=GetVersionEx(&vi);
	
	FALSE_EXCEPTION(fres,"win_platform");

	if(vi.dwMajorVersion >= ma && vi.dwMinorVersion >= mi)
	{
		result = ETRUE;
	}
	return(result);
}

/* timer */

FRTAPI void win_timer_start(E_HWND win, EINT id, EINT delay)
{
	UINT fres;

	fres=SetTimer(E2W_HWND(win), E2W_INT(id), E2W_INT(delay), NULL);
	ZERO_EXCEPTION(fres,"win_timer_start");
}

FRTAPI void win_timer_stop(E_HWND win, EINT id)
{
	BOOL fres;

	fres=KillTimer(E2W_HWND(win), E2W_INT(id));
	FALSE_EXCEPTION(fres,"win_timer_stop");
}

/* registry */

FRTAPI E_HDLE win_reg_open(EINT base, ESTRING name)
{
	HKEY result;
	
	if(RegOpenKeyEx((HKEY) (0x80000000 | E2W_INT(base)), E2W_STRING(name), 0,
			KEY_ALL_ACCESS,&result)!=ERROR_SUCCESS)
		result = NULL;
	return(W2E_HDLE(result));
}

FRTAPI E_HDLE win_reg_create(EINT base, ESTRING name)
{
	HKEY result;
	DWORD new;
	
	if(RegCreateKeyEx((HKEY) (0x80000000 | E2W_INT(base)), E2W_STRING(name), 0,
			TEXT("Fenestra Data"), REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS,NULL,
			&result,&new)!=ERROR_SUCCESS)
		result = NULL;
	else
	{
		if (new == REG_CREATED_NEW_KEY)
			fenestra_set_bool(ETRUE);
		else
			fenestra_set_bool(EFALSE);
	}
	return(W2E_HDLE(result));
}

FRTAPI void win_reg_close(E_HDLE keyid)
{
	LONG fres;
	
	fres=RegCloseKey(E2W_HDLE(keyid));
	ASSERT_EXCEPTION(fres == ERROR_SUCCESS, "win_reg_close");
}

FRTAPI EBOOL win_reg_delete(E_HDLE keyid, ESTRING subkey)
{
	EBOOL result;

	result = EFALSE;

	if(RegDeleteKey(E2W_HDLE(keyid),E2W_STRING(subkey))==ERROR_SUCCESS)
		result = ETRUE;

	return(result);
}

FRTAPI EBOOL win_reg_values(E_HDLE keyid, EINT index)
{
	EBOOL result;
	DWORD value_size;
	LPTSTR value;

	result = EFALSE;
	fenestra_void_string();
	value_size = 0;

	if(RegQueryInfoKey(E2W_HDLE(keyid),NULL,NULL,NULL,NULL,NULL,
		NULL,NULL,&value_size,NULL,NULL,NULL)==ERROR_SUCCESS)
	{
		value=fenestra_malloc((value_size+1)*sizeof(TCHAR));
		if(RegEnumValue(E2W_HDLE(keyid),E2W_INT(index),value,&value_size,
			NULL,NULL,NULL,NULL)==ERROR_SUCCESS)
		{
			fenestra_pass_string(value); /* will free */
			result = ETRUE;
		}
		else
		{
			fenestra_free(value);
		}
	}
	return(result);
}

FRTAPI EBOOL win_reg_subkeys(E_HDLE keyid, EINT index)
{
	EBOOL result;
	DWORD value_size;
	LPTSTR value;

	result = EFALSE;
	fenestra_void_string();
	value_size = 0;

	if(RegQueryInfoKey(E2W_HDLE(keyid),NULL,NULL,NULL,NULL,&value_size,
		NULL,NULL,NULL,NULL,NULL,NULL)==ERROR_SUCCESS)
	{
		value=fenestra_malloc((value_size+1)*sizeof(TCHAR));
		if(RegEnumKey(E2W_HDLE(keyid),E2W_INT(index),value,value_size)==ERROR_SUCCESS)
		{
			fenestra_pass_string(value); /* will free */
			result = ETRUE;
		}
		else
		{
			fenestra_free(value);
		}
	}
	return(result);
}

FRTAPI EINT win_reg_type(E_HDLE keyid, ESTRING label)
{
	DWORD type;
	
	if(RegQueryValueEx(E2W_HDLE(keyid),E2W_STRING(label),NULL,&type,
			NULL,NULL)!=ERROR_SUCCESS)
		type=0;
	
	return(W2E_INT(type));
}

FRTAPI EINT win_reg_getint(E_HDLE keyid, ESTRING label)
{
	DWORD type,result;
	DWORD size;
	LONG fres;

	size = sizeof(DWORD);
	fres=RegQueryValueEx(E2W_HDLE(keyid),E2W_STRING(label),NULL,&type,
			(LPBYTE) &result, &size);

	ASSERT_EXCEPTION(fres == ERROR_SUCCESS, "win_reg_getint.1");
	ASSERT_EXCEPTION(type == REG_DWORD, "win_reg_getint.2");

	return(W2E_INT(result));
}

FRTAPI void win_reg_getstring(E_HDLE keyid, ESTRING label)
{
	LPTSTR wintext;
	DWORD type,textlen;
	LONG fres;

	/* default result */
	fenestra_void_string();

	/* get lenght (binary, incl. 0, not characters) */
	fres=RegQueryValueEx(E2W_HDLE(keyid),E2W_STRING(label),NULL,&type,NULL,&textlen);
	ASSERT_EXCEPTION(fres == ERROR_SUCCESS, "win_reg_getstring.1");

	/* get actual string */
	if(textlen > 0)
	{
		wintext=fenestra_malloc(textlen+2);
		wintext[0]=((TCHAR) 0);
		
		fres=RegQueryValueEx(E2W_HDLE(keyid),E2W_STRING(label),NULL,&type,(LPBYTE) wintext,&textlen);
		ASSERT_EXCEPTION(fres==ERROR_SUCCESS, "win_reg_getstring.2");

		fenestra_pass_string(wintext); /* will free */
	}
}

FRTAPI void win_reg_putint(E_HDLE keyid, ESTRING label, EINT value)
{
	DWORD data;
	LONG fres;
	
	data = E2W_INT(value);
	fres=RegSetValueEx(E2W_HDLE(keyid),E2W_STRING(label),0,
			REG_DWORD,(BYTE *)&data,sizeof(DWORD));
	ASSERT_EXCEPTION(fres == ERROR_SUCCESS, "win_reg_putint");
}

FRTAPI void win_reg_putstring(E_HDLE keyid, ESTRING label, ESTRING value)
{
	LPTSTR data;
	LONG fres;

	data = E2W_STRING(value);
	fres=RegSetValueEx(E2W_HDLE(keyid),E2W_STRING(label),0,
			REG_SZ,(BYTE *) data, (lstrlen(data) + 1) * sizeof(TCHAR) );
	ASSERT_EXCEPTION(fres == ERROR_SUCCESS, "win_reg_putstring");
}

/* memory mapped file */

FRTAPI EPOINTER win_mmap_file(E_HDLE file_handle)
{
	LPVOID result;
	HANDLE filemap;

	fenestra_set_bool(EFALSE);
	result=NULL; /* error */

	filemap=CreateFileMapping(E2W_HDLE(file_handle),NULL,PAGE_READONLY,0,0,NULL);
	if(filemap!=NULL) 
	{
		result=MapViewOfFile(filemap,FILE_MAP_READ,0,0,0);
		/* error: result==NULL */
		
		CloseHandle(filemap); /* but map still alive */
	}
	
	if(result!=NULL) 
	{
		fenestra_set_bool(ETRUE);
	}
	return(W2E_POINTER(result));
}

FRTAPI void win_mmap_free(EPOINTER mmap_ptr)
{
	BOOL res;

	res=UnmapViewOfFile(E2W_POINTER(mmap_ptr));
	FALSE_EXCEPTION(res,"win_mmap_free");
}

/* synchronisation objects */

FRTAPI EBOOL win_synch_wait(E_HDLE handle, EINT delay_ms)
{
	DWORD res;
	EBOOL result;

	result = EFALSE;
	fenestra_set_bool(EFALSE);
	
	res=WaitForSingleObject(E2W_HDLE(handle),E2W_INT(delay_ms));
	ASSERT_EXCEPTION(res!=WAIT_FAILED,"win_synch_wait");
	
	if (res==WAIT_OBJECT_0)
	{
		result = ETRUE;
	}
	if (res==WAIT_ABANDONED)
	{
		result = ETRUE;
		fenestra_set_bool(ETRUE);
	}

	return(result);
}

FRTAPI EINT win_synch_mwait(EINTARRAY obj_handles, EINT obj_count, EINT delay_ms)
{
	DWORD res;
	DWORD cnt;
	EINT result;
	
	cnt = E2W_INT(obj_count);

	result = W2E_INT(-1);
	fenestra_set_bool(EFALSE);

	res=WaitForMultipleObjects(cnt,(HANDLE *) E2W_INTARRAY(obj_handles),FALSE,E2W_INT(delay_ms));
	if ((res >= WAIT_OBJECT_0) && (res < (WAIT_OBJECT_0 + cnt)))
	{
		result = W2E_INT(res - WAIT_OBJECT_0);
	}
	else
	{
		if((res >= WAIT_ABANDONED_0) && (res < (WAIT_ABANDONED_0 + cnt)))
		{
			result = W2E_INT(res - WAIT_ABANDONED_0);
			fenestra_set_bool(ETRUE);
		}
		else
		{
			ASSERT_EXCEPTION(res = WAIT_TIMEOUT,"win_synch_mwait"); /* fixme: too strict? */
			/* result = -1 */
		}
	}
	return(result);
}

FRTAPI E_HDLE win_event_create(ESTRING name, EBOOL manual_reset)
{
	HANDLE result;
	LPTSTR wname;

	/* check void */
	wname=E2W_STRING(name);
	if(lstrlen(wname)==0)
		wname=NULL;

	result = CreateEvent(NULL,E2W_BOOL(manual_reset),FALSE,wname);
	return(W2E_HDLE(result));
}

FRTAPI E_HDLE win_event_open(ESTRING name)
{
	HANDLE result;

	result = OpenEvent(EVENT_ALL_ACCESS,FALSE,E2W_STRING(name));
	return(W2E_HDLE(result));
}

FRTAPI void win_event_set(E_HDLE handle)
{
	BOOL res;

	res=SetEvent(E2W_HDLE(handle));
	FALSE_EXCEPTION(res,"win_event_set");
}

FRTAPI void win_event_reset(E_HDLE handle)
{
	BOOL res;

	res=ResetEvent(E2W_HDLE(handle));
	FALSE_EXCEPTION(res,"win_event_reset");
}

FRTAPI E_HDLE win_mutex_create(ESTRING name)
{
	HANDLE handle;

	handle = CreateMutex(NULL,FALSE,E2W_STRING(name));
	return(W2E_HDLE(handle));
}

FRTAPI E_HDLE win_mutex_open(ESTRING name)
{
	HANDLE handle;

	handle = OpenMutex(MUTEX_ALL_ACCESS,FALSE,E2W_STRING(name));
	return(W2E_HDLE(handle));
}

FRTAPI void win_mutex_release(E_HDLE handle)
{
	BOOL res;

	res=ReleaseMutex(E2W_HDLE(handle));
	FALSE_EXCEPTION(res,"win_mutex_release");
}

FRTAPI E_HDLE win_sema_create(ESTRING name, EINT max_count)
{
	HANDLE handle;
	LPTSTR wname;

	wname=E2W_STRING(name);
	if (lstrlen(wname)==0)
		wname=NULL;

	handle=CreateSemaphore(NULL,E2W_INT(max_count),E2W_INT(max_count),wname);
	return(W2E_HDLE(handle));
}

FRTAPI E_HDLE win_sema_open(ESTRING name)
{
	HANDLE handle;

	handle=OpenSemaphore(SEMAPHORE_ALL_ACCESS,FALSE,E2W_STRING(name));
	return(W2E_HDLE(handle));
}

FRTAPI void win_sema_release(E_HDLE handle)
{
	BOOL res;

	res=ReleaseSemaphore(E2W_HDLE(handle),1,NULL);
	FALSE_EXCEPTION(res,"win_sema_release");
}

/* TCP/IP */

FRTAPI void win_tcp_init(void)
{
	int err;
	WSADATA sock_data;

	/* check using invalid handle ok for socket */
	ASSERT_EXCEPTION(((HANDLE)INVALID_SOCKET) == INVALID_HANDLE_VALUE,"win_tcp_init.1");

	err = WSAStartup(MAKEWORD(1,1), &sock_data); 
	ASSERT_EXCEPTION(err == 0, "win_tcp_init.2");
}

FRTAPI void win_tcp_finish(void)
{
	int err;

	err = WSACleanup();
	ASSERT_EXCEPTION(err == 0, "win_tcp_finish");
}

FRTAPI E_HDLE win_tcp_connect(EINT ip1, EINT ip2, EINT ip3, EINT ip4, EINT protocol)
{
	SOCKET sock;
	HANDLE result;
	struct sockaddr_in sock_addr;
	struct in_addr addr;
	int err;

	sock = socket(PF_INET,SOCK_STREAM,0);
	if (sock != INVALID_SOCKET)
	{
		/* IP address */
		addr.S_un.S_un_b.s_b1 = (u_char) E2W_INT(ip1);
		addr.S_un.S_un_b.s_b2 = (u_char) E2W_INT(ip2);
		addr.S_un.S_un_b.s_b3 = (u_char) E2W_INT(ip3);
		addr.S_un.S_un_b.s_b4 = (u_char) E2W_INT(ip4);

		/* init address struct */
		fenestra_memset(&sock_addr,sizeof(struct sockaddr_in));
		sock_addr.sin_family = AF_INET; /* 2 : TCP/IP */
		sock_addr.sin_port = (u_short) htons((u_short) E2W_INT(protocol)); /* protocol */
		sock_addr.sin_addr.S_un.S_addr = htonl(addr.S_un.S_addr);

		err = connect(sock,(struct sockaddr *) &sock_addr,sizeof(struct sockaddr_in));
		if (err == 0)
		{
			result = (HANDLE) sock;
		}
		
	}
	result = INVALID_HANDLE_VALUE;

	return(W2E_HDLE(result));
}

FRTAPI EBOOL win_tcp_byname(ESTRING name)
{
	EBOOL result;
	struct hostent *res;
	struct in_addr addr;

	result = EFALSE;

	res=gethostbyname(W2C_STRING(E2W_STRING(name)));
	if(res!=NULL)
	{
		addr.S_un.S_addr = ntohl( *((u_long *) res->h_addr_list[0]));
		fenestra_set_int4(E2W_INT(addr.S_un.S_un_b.s_b1),
				E2W_INT(addr.S_un.S_un_b.s_b2),
				E2W_INT(addr.S_un.S_un_b.s_b3),
				E2W_INT(addr.S_un.S_un_b.s_b4));
		
		result = ETRUE;
	}

	return(result);
}

/* FRTAPI void win_tcp_close(E_HDLE handle) { closesocket(E2W_HDLE(handle)); } */

/* file */

FRTAPI E_HDLE win_file_open(ESTRING name, EINT create, EBOOL read, EBOOL write, EBOOL sharer, EBOOL sharew, EBOOL random)
{
	DWORD rflag = 0;
	DWORD sflag = 0;
	DWORD attflag = FILE_ATTRIBUTE_NORMAL;
	HANDLE result;

	if (read == ETRUE) rflag |= GENERIC_READ;
	if (write == ETRUE) rflag |= GENERIC_WRITE;
	if (sharer == ETRUE) sflag |= FILE_SHARE_READ;
	if (sharew == ETRUE) sflag |= FILE_SHARE_WRITE;
	if (random == ETRUE) 
		attflag |= FILE_FLAG_RANDOM_ACCESS;
	else
		attflag |= FILE_FLAG_SEQUENTIAL_SCAN;
	
	result = CreateFile(E2W_STRING(name),rflag,sflag,NULL,
		E2W_INT(create),attflag,NULL);

	return(W2E_HDLE(result));
}

FRTAPI void win_file_close(E_HDLE fileid)
{
	BOOL fres;

	fres=CloseHandle(E2W_HDLE(fileid));
	FALSE_EXCEPTION(fres, "win_file_close");
}

FRTAPI EINT win_file_read(E_HDLE fileid, EBYTEARRAY buf, EINT bufsize)
{
	DWORD read;
	BOOL wresult;

	DWORD size;
	BYTE *ptr;

	size = E2W_INT(bufsize);
	ptr = E2W_BYTEARRAY(buf);

	wresult = ReadFile(E2W_HDLE(fileid),
					ptr,
					size,
					&read,NULL);
	
	FALSE_EXCEPTION(wresult,"win_file_readbyte"); /* fixme: good idea? */
   
	return(W2E_INT(read));
}

FRTAPI EINT win_file_write(E_HDLE fileid, EBYTEARRAY ptr, EINT len)
{
	DWORD written;
	BOOL wresult;

	wresult = WriteFile(E2W_HDLE(fileid),E2W_BYTEARRAY(ptr),E2W_INT(len),&written,NULL);
	FALSE_EXCEPTION(wresult, "win_file_write"); /* fixme: use written = 0? */

	return(W2E_INT(written));
}

FRTAPI EBOOL win_file_asread(E_HDLE fileid, EBYTEARRAY buf, EINT bufsize, EPOINTER overlapped)
{
	DWORD read;
	BOOL wresult;

	fenestra_set_bool(EFALSE);

	wresult = ReadFile(E2W_HDLE(fileid),
					E2W_BYTEARRAY(buf),
					(DWORD) E2W_INT(bufsize),
					&read,E2W_POINTER(overlapped));
	
	if((wresult == FALSE) && (GetLastError() == ERROR_IO_PENDING))
	{
		fenestra_set_bool(ETRUE);
	}

	return(W2E_BOOL(wresult));
}

FRTAPI EINT win_file_aswrite(E_HDLE fileid, EBYTEARRAY ptr, EINT len, EPOINTER overlapped)
{
	DWORD written;
	BOOL wresult;

	fenestra_set_bool(EFALSE);

	wresult = WriteFile(E2W_HDLE(fileid),
			E2W_BYTEARRAY(ptr), (DWORD) E2W_INT(len),
			&written,E2W_POINTER(overlapped));

	if((wresult == FALSE) && (GetLastError() == ERROR_IO_PENDING))
	{
		fenestra_set_bool(ETRUE);
	}

	return(W2E_BOOL(wresult));
}

FRTAPI EINT win_file_asok(E_HDLE fileid, EPOINTER overlapped)
{
	DWORD result;
	BOOL res;

	result = 0;
	res=GetOverlappedResult(E2W_HDLE(fileid), E2W_POINTER(overlapped), &result, FALSE);
	if(res==FALSE)
	{
		result = 0; 
	}

	return(W2E_INT(result));
}

FRTAPI EBOOL win_file_writebyte(E_HDLE fileid, EINT byte)
{
	BYTE wbyte;
	DWORD written;
	BOOL wresult;

	wbyte = (BYTE) byte;

	wresult = WriteFile(E2W_HDLE(fileid),&wbyte,1,&written,NULL); /* good idea? */
	FALSE_EXCEPTION(wresult, "win_file_writebyte");
	
	if (written!=1)
		return(EFALSE);
	return(ETRUE);
}

FRTAPI EINT win_file_ptr(E_HDLE fileid)
{
	DWORD dw;
	EINT result;

	dw = SetFilePointer(E2W_HDLE(fileid),0,NULL,FILE_CURRENT);
	if (dw == 0xFFFFFFFF)
		result = W2E_INT(-1);
	else
		result = W2E_INT(dw);
	return(result);
}

FRTAPI EBOOL win_file_setptr(E_HDLE fileid, EINT index)
{
	EBOOL result;
	
	result = ETRUE;
	if(SetFilePointer(E2W_HDLE(fileid),E2W_INT(index),NULL,FILE_BEGIN)==0xFFFFFFFF)
		result = EFALSE;	
	return(result);
}

FRTAPI EINT win_file_size(E_HDLE fileid)
{
	DWORD dw;
	EINT result;

	dw = GetFileSize(E2W_HDLE(fileid),NULL);
	if (dw == 0xFFFFFFFF)
		result = W2E_INT(-1);
	else
		result = W2E_INT(dw);
	return(result);
} 

/* file system */

FRTAPI EINT win_file_dspace(ESTRING drive)
{
	EINT result;
	LPTSTR drivename;
	DWORD sec_cluster,byte_sector,free_cluster,total_cluster;

	drivename = E2W_STRING(drive);
	if (*drivename==0)
		drivename=NULL;
	
	if(GetDiskFreeSpace(drivename,&sec_cluster,&byte_sector,
	   		&free_cluster,&total_cluster)==FALSE)
		result = W2E_INT(-1);

	result = W2E_INT((MulDiv (sec_cluster * byte_sector, free_cluster, 1024)));
	return(result);
}

FRTAPI EBOOL win_file_attrib(ESTRING fname, EINT flag)
{
	DWORD mask;
	DWORD fattrib;
	EBOOL result;

	result = EFALSE;
	mask = E2W_INT(flag);

	fattrib = GetFileAttributes(E2W_STRING(fname));
	if((fattrib != 0xFFFFFFFF) && (fattrib & mask))
		result = ETRUE;
	return(result);
}

FRTAPI EBOOL win_file_mkdir(ESTRING pname)
{
	return(W2E_BOOL(CreateDirectory(E2W_STRING(pname),NULL)));
}

FRTAPI EBOOL win_file_rmdir(ESTRING pname)
{
	return(W2E_BOOL(RemoveDirectory(E2W_STRING(pname))));
}

FRTAPI void win_file_curdir(void)
{
	DWORD dirlen;
	LPTSTR curdir;

	curdir=fenestra_malloc((MAX_PATH+2) * sizeof(TCHAR));

	dirlen=GetCurrentDirectory(MAX_PATH+1,curdir);

	if(dirlen >= MAX_PATH)
	{ /* too small, realloc new buffer */
		fenestra_free(curdir);
		curdir=fenestra_malloc((dirlen+1)*sizeof(TCHAR));
		
		dirlen=GetCurrentDirectory(dirlen,curdir);
	}
	if(dirlen <= 0) {
		curdir[0]=((TCHAR) 0);
	}
	
	fenestra_pass_string(curdir); /* will free */
}

FRTAPI void win_file_tmpdir(void)
{
	DWORD dirlen;
	LPTSTR curdir;

	curdir=fenestra_malloc((MAX_PATH+2) * sizeof(TCHAR));

	dirlen=GetTempPath(MAX_PATH+1,curdir);

	if(dirlen >= MAX_PATH)
	{ /* too small, realloc new buffer */
		fenestra_free(curdir);
		curdir=fenestra_malloc((dirlen+1)*sizeof(TCHAR));
		
		dirlen=GetTempPath(dirlen,curdir);
	}
	if(dirlen <= 0) {
		curdir[0]=((TCHAR) 0);
	}

	fenestra_pass_string(curdir); /* will free */
}

FRTAPI EBOOL win_file_setcurdir(ESTRING newdir)
{
	return(W2E_BOOL(SetCurrentDirectory(E2W_STRING(newdir))));	
}

FRTAPI void win_file_ffirst(ESTRING pattern)
{
	HANDLE fhandle;
	WIN32_FIND_DATA fd;

	fenestra_void_string();

	fhandle = FindFirstFile(E2W_STRING(pattern),&fd);
	fenestra_set_int(W2E_INT(fhandle)); /* fixme: use handle conversion? */
	if (fhandle != INVALID_HANDLE_VALUE) {
		fenestra_copy_string((LPTSTR) &(fd.cFileName));
	}
}

FRTAPI void win_file_fnext(E_HDLE handle)
{
	BOOL resbool;
	WIN32_FIND_DATA fd;

	fenestra_void_string();

	resbool =FindNextFile(E2W_HDLE(handle),&fd);
	fenestra_set_bool(W2E_BOOL(resbool));

	if (resbool==TRUE) {
		fenestra_copy_string((LPTSTR) &(fd.cFileName));
	}
}

FRTAPI void win_file_fclose(E_HDLE handle)
{
	BOOL fres;

	fres=FindClose(E2W_HDLE(handle));
	FALSE_EXCEPTION(fres, "win_file_fclose");
}

FRTAPI EBOOL win_file_exists(ESTRING fname)
{
	HANDLE hdle;
	EBOOL result;
	WIN32_FIND_DATA fd;
	
	hdle=FindFirstFile(E2W_STRING(fname),&fd);
	
	if (hdle==INVALID_HANDLE_VALUE)
		result = EFALSE;
	else
	{
		FindClose(hdle);
		result = ETRUE;	
	}
	
	return(result);
}

FRTAPI EBOOL win_file_remove(ESTRING fname)
{
	return(W2E_BOOL(DeleteFile(E2W_STRING(fname))));
}

FRTAPI EBOOL win_file_copy(ESTRING orig, ESTRING dest)
{
	return(W2E_BOOL(CopyFile(E2W_STRING(orig),E2W_STRING(dest),TRUE)));
}

FRTAPI EBOOL win_file_move(ESTRING orig, ESTRING dest)
{
	return(W2E_BOOL(MoveFile(E2W_STRING(orig),E2W_STRING(dest))));
}

FRTAPI EBOOL win_file_prop(ESTRING fname, EFINDDATA fdata)
{
	HANDLE hdle;
	EBOOL result=EFALSE;

	hdle=FindFirstFile(E2W_STRING(fname),E2W_FINDDATA(fdata));
	if (hdle != INVALID_HANDLE_VALUE)
	{
		result = ETRUE;
		FindClose(hdle);
	}
	return(result);
}

FRTAPI void win_file_convtime(EFILETIME ft, ESYSTIME st)
{
	BOOL fres;

	fres=FileTimeToSystemTime(E2W_FILETIME(ft),E2W_SYSTIME(st));
	FALSE_EXCEPTION(fres,"win_file_convtime");	
}

FRTAPI EINT win_last_error(void)
{
	return(W2E_INT(GetLastError()));
}

/* Time */

FRTAPI EINT win_time(EINT zone, EINT info)
{
	SYSTEMTIME wtime;
	WORD result;

	if (E2W_INT(zone) == 1) /* Filo_local */
		GetLocalTime(&wtime);
	else  /* Filo_gmt (0) */
		GetSystemTime(&wtime);

	switch (E2W_INT(info))
	{
		case 10:	result = wtime.wYear;	break; /* Filo_year */
		case 11:	result = wtime.wMonth;	break; /* Filo_month */
		case 12:	result = wtime.wDay;	break; /* Filo_day */
		case 13:	result = wtime.wHour;	break; /* Filo_hour */
		case 14:	result = wtime.wMinute;	break; /* Filo_minute */
		case 15:	result = wtime.wSecond;	break; /* Filo_second */
	}

	return(W2E_INT(result));
}

FRTAPI EINT win_timezone(void)
{
	TIME_ZONE_INFORMATION tz;
	DWORD fres;

	fres=GetTimeZoneInformation(&tz);
	ASSERT_EXCEPTION(fres!=0xFFFFFFFF, "win_timezone");
	
	return(W2E_INT(tz.Bias));
} 

FRTAPI EINT win_tick(void)
{
	DWORD result;

	result = GetTickCount();
	result &= 0x7FFFFFFF; /* positive eiffel integer, it looks like if the tick 
							 counter is wraping around twice as often (every 25 days) */
	return(W2E_INT(result));
}

/* System services */

FRTAPI EBOOL win_execute(ESTRING command)
{
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	BOOL result;
		
	DWORD flags = CREATE_NEW_CONSOLE;  /* fire & forget */

	fenestra_memset(&si, sizeof(STARTUPINFO));
	si.cb = sizeof(STARTUPINFO);
	/* use default values otherwise */
		
	result=CreateProcess(NULL,	/* image file name */
		E2W_STRING(command),	/* command line (including program name) */
		NULL,  					/* security for process */
		NULL,					/* security for main thread */
		FALSE,					/* new process inherits handles? */
		flags,					/* creation flags */
		NULL,					/* environment */
		NULL,  					/* use current process' directory */
		&si,					/* STARTUPINFO structure */
		&pi);
	
	if(result != FALSE) /* PROCESSINFORMATION structure */
	{
		CloseHandle(pi.hThread);
		CloseHandle(pi.hProcess);  /* close process handle or it won't die */
	}

	return(W2E_BOOL(result));
}

FRTAPI EINT win_getpid(void)
{
	return(W2E_INT(GetCurrentProcessId()));
}

FRTAPI void win_beep(EINT beeptype)
{
	BOOL fres;

	fres=MessageBeep(E2W_INT(beeptype) == 0 ? 0xFFFFFFFF : E2W_INT(beeptype));
	FALSE_EXCEPTION(fres,"win_beep"); 
}

FRTAPI void win_sleep(EINT ms)
{
	Sleep(E2W_INT(ms));
}

FRTAPI void win_debugout(ESTRING str)
{
	OutputDebugString(E2W_STRING(str));
}

FRTAPI void win_rsrcstring(EINT id)
{
	LPTSTR rstr;

	rstr = fenestra_malloc((RSRC_STRING+1) * sizeof(TCHAR));
	
	if(LoadString(PROGHINST,E2W_INT(id),rstr,RSRC_STRING) == 0) {
		rstr[0] = ((TCHAR) 0);
	}
	
	fenestra_pass_string(rstr); /* will free */
}

/* Multimedia */

FRTAPI EINT win_mci_send(ESTRING str)
{
	return(W2E_INT(mciSendString(E2W_STRING(str),(LPTSTR) NULL,0,(HANDLE) NULL)));
}

FRTAPI void win_mci_error(EINT code)
{
	LPTSTR errstr;

	errstr = fenestra_malloc((MCI_ERROR_STRING+1)*sizeof(TCHAR));
	
	if(mciGetErrorString(E2W_INT(code),errstr,MCI_ERROR_STRING)==FALSE) {
		errstr[0] = ((TCHAR) 0);
	}
	
	fenestra_pass_string(errstr); /* will free */
}

/**/